# -*- coding: utf-8 -*-
"""
This is a the source file that contains the Rule, RuleBase and MasterRuleBase classes to perform fuzzy inference.

"""
import abc

import numpy as np
try:
    from . import fuzzy_sets as fs
    from . import centroid
except ImportError:
    import fuzzy_sets as fs
    import centroid


def compute_antecedents_memberships(antecedents: list[fs.fuzzyVariable], x: np.array) -> list[dict]:
        '''
        Returns a list of of dictionaries that contains the memberships for each x value to the ith antecedents, nth linguistic variable.
        x must be a vector (only one sample)

        :param x: vector with the values of the inputs.
        :return: a list with the antecedent truth values for each one. Each list is comprised of a list with n elements, where n is the number of linguistic variables in each variable.
        '''
        x = np.array(x)
        cache_antecedent_memberships = []

        for ix, antecedent in enumerate(antecedents):
            cache_antecedent_memberships.append(
                antecedent.compute_memberships(x[:, ix]))

        return cache_antecedent_memberships

        
            
class RuleError(Exception):
    '''
    Exception raised when a rule is not well defined.
    '''

    def __init__(self, message: str) -> None:
        '''
        Constructor of the RuleError class.

        :param message: message of the error.
        '''
        super().__init__(message)

def _myprod(x: np.array, y: np.array) -> np.array:
    '''
    Proxy function to change the product operation interface
    '''
    return x*y


class Rule():
    '''
    Class of Rule designed to work with one single rule. It contains the whole inference functionally in itself.
    '''

    def __init__(self, antecedents: list[fs.FS], consequent: fs.FS) -> None:
        '''
        Creates a rule with the given antecedents and consequent.

        :param antecedents: list of fuzzy sets.
        :param consequent: fuzzy set.
        '''
        self.antecedents = antecedents
        self.consequent = consequent

    def membership(self, x: np.array, tnorm=_myprod) -> np.array:
        '''
        Computes the membership of one input to the antecedents of the rule.

        :param x: input to compute the membership.
        :param tnorm: t-norm to use in the inference process.
        '''
        for ix, antecedent in enumerate(self.antecedents):
            if ix == 0:
                res = antecedent.membership(x)
            else:
                res = tnorm(res, antecedent.membership(x))

        return res

    def consequent_centroid(self) -> np.array:
        '''
        Returns the centroid of the consequent using a Karnik and Mendel algorithm.
        '''
        try:
            return self.centroid_consequent
        except AttributeError:
            consequent_domain = self.consequent.domain
            domain_linspace = np.arange(
                consequent_domain[0], consequent_domain[1], 0.05)
            consequent_memberships = self.consequent.membership(
                domain_linspace)

            self.centroid_consequent = centroid.compute_centroid_fs(
                domain_linspace, consequent_memberships)

            return self.centroid_consequent


class RuleSimple():
    '''
    Class designed to represent rules in its simplest form to optimize the computation in a rulebase.

    It indicates the antecedent values using a vector coded in this way:
        ith entry: -1 means this variable is not used. 0-N indicates that the variable linguistic used for the ith input is that one.
    '''

    def __init__(self, antecedents: list[int], consequent: int=0) -> None:
        '''
        Creates a rule with the given antecedents and consequent.

        :param antecedents: list of integers indicating the linguistic variable used for each input.
        :param consequent: integer indicating the linguistic variable used for the consequent.
        '''
        self.antecedents = list(map(int, antecedents))
        self.consequent = int(consequent)
        self.score = 1.0


    def __getitem__(self, ix):
        '''
        Returns the antecedent value for the given index.

        :param ix: index of the antecedent to return.
        '''
        return self.antecedents[ix]


    def __setitem__(self, ix, value):
        '''
        Sets the antecedent value for the given index.

        :param ix: index of the antecedent to set.
        :param value: value to set.
        '''
        self.antecedents[ix] = value
    

    def __str__(self):
        '''
        Returns a string representation of the rule.
        '''
        return 'Rule: antecedents: ' + str(self.antecedents) + ' consequent: ' + str(self.consequent)
    

    def __len__(self):
        '''
        Returns the number of antecedents in the rule.
        '''
        return len(self.antecedents)
    

    def __eq__(self, other):
        '''
        Returns True if the two rules are equal.
        '''
        return self.antecedents == other.antecedents and self.consequent == other.consequent
    

    def __hash__(self):
        '''
        Returns the hash of the rule.
        '''
        return hash(str(self))


class RuleBase():
    '''
    Class optimized to work with multiple rules at the same time. Right now supports only one consequent. (Solution: use one rulebase per consequent to study)
    '''

    def __init__(self, antecedents: list[fs.fuzzyVariable], rules: list[RuleSimple], consequent: fs.fuzzyVariable = None, tnorm=np.prod) -> None:
        '''
        Creates a rulebase with the given antecedents, rules and consequent.

        :param antecedents: list of fuzzy sets.
        :param rules: list of rules.
        :param consequent: fuzzy set.
        :param tnorm: t-norm to use in the inference process.
        '''
        rules = self.delete_rule_duplicates(rules)
        self.rules = rules
        self.antecedents = antecedents
        self.consequent = consequent
        self.tnorm = tnorm

        if consequent is not None:
            self.consequent_centroids = np.zeros(
                (len(consequent.linguistic_variable_names()), 2))
            for ix, vl_consequent in enumerate(consequent.linguistic_variables):
                consequent_domain = vl_consequent.domain
                domain_linspace = np.arange(
                    consequent_domain[0], consequent_domain[1], 0.05)
                consequent_memberships = vl_consequent.membership(domain_linspace)

                self.consequent_centroids[ix, :] = centroid.compute_centroid_iv(
                    domain_linspace, consequent_memberships)

            self.consequent_centroids_rules = np.zeros((len(self.rules), 2))
            for ix, rule in enumerate(self.rules):
                consequent_ix = rule.consequent
                self.consequent_centroids_rules[ix] = self.consequent_centroids[consequent_ix]

        self.delete_duplicates()


    def get_rules(self) -> list[RuleSimple]:
        '''
        Returns the list of rules in the rulebase.
        '''
        return self.rules


    def add_rule(self, new_rule: RuleSimple):
        '''
        Adds a new rule to the rulebase.
        :param new_rule: rule to add.
        '''
        self.rules.append(new_rule)


    def add_rules(self, new_rules: list[RuleSimple]):
        '''
        Adds a list of new rules to the rulebase.

        :param new_rules: list of rules to add.
        '''
        self.rules += new_rules


    def remove_rule(self, ix: int) -> None:
        '''
        Removes the rule in the given index.
        :param ix: index of the rule to remove.
        '''
        del self.rules[ix]


    def remove_rules(self, delete_list: list[int]) -> None:
        '''
        Removes the rules in the given list of indexes.

        :param delete_list: list of indexes of the rules to remove.
        '''
        self.rules = [rule for ix, rule in enumerate(
            self.rules) if ix not in delete_list]


    def get_rulebase_matrix(self):
        '''
        Returns a matrix with the antecedents values for each rule.
        '''
        res = np.zeros((len(self.rules), len(self.antecedents)))

        for ix, rule in enumerate(self.rules):
            res[ix] = rule

        return res


    def get_scores(self):
        '''
        Returns an array with the dominance score for each rule.
        (Must been already computed by an evalRule object)

        '''
        res = np.zeros((len(self.rules, )))
        for ix, rule in enumerate(self.rules):
            res[ix] = rule.score

        return res


    def delete_rule_duplicates(self, list_rules:list[RuleSimple]):
        # Delete the rules that are duplicated in the rule list
        unique = {}
        for ix, rule in enumerate(list_rules):
            try:
                unique[rule]
            except KeyError:
                unique[rule] = ix

        new_list = [list_rules[x] for x in unique.values()]
        
        return new_list
        

    def compute_antecedents_memberships(self, x: np.array) -> list[dict]:
        '''
        Returns a list of of dictionaries that contains the memberships for each x value to the ith antecedents, nth linguistic variable.
        x must be a vector (only one sample)

        :param x: vector with the values of the inputs.
        :return: a list with the antecedent truth values for each one. Each list is comprised of a list with n elements, where n is the number of linguistic variables in each variable.
        '''
        if len(self.rules) > 0:
            cache_antecedent_memberships = []

            for ix, antecedent in enumerate(self.antecedents):
                cache_antecedent_memberships.append(
                    antecedent.compute_memberships(x[:, ix]))

            return cache_antecedent_memberships

        else:
            if self.fuzzy_type() == fs.FUZZY_SETS.t1:
                return [np.zeros((x.shape[0], 1))]
            elif self.fuzzy_type() == fs.FUZZY_SETS.t2:
                return [np.zeros((x.shape[0], 1, 2))]
            elif self.fuzzy_type() == fs.FUZZY_SETS.gt2:
                return [np.zeros((x.shape[0], len(self.alpha_cuts), 2))]


    def compute_rule_antecedent_memberships(self, x: np.array, scaled=False, antecedents_memberships:list[np.array]=None) -> np.array:
        '''
        Computes the antecedent truth value of an input array.

        Return an array in shape samples x rules (x 2) (last is iv dimension)

        :param x: array with the values of the inputs.
        :param scaled: if True, the memberships are scaled according to their sums for each sample.
        :return: array with the memberships of the antecedents for each rule.
        '''
        if self.fuzzy_type() == fs.FUZZY_SETS.t2:
            res = np.zeros((x.shape[0], len(self.rules), 2))
        elif self.fuzzy_type() == fs.FUZZY_SETS.t1:
            res = np.zeros((x.shape[0], len(self.rules), ))
        elif self.fuzzy_type() == fs.FUZZY_SETS.gt2:
            res = np.zeros(
                (x.shape[0], len(self.rules), len(self.alpha_cuts), 2))

        if antecedents_memberships is None:
            antecedents_memberships = self.compute_antecedents_memberships(x)
        
        for jx, rule in enumerate(self.rules):
            rule_antecedents = rule.antecedents

            if self.fuzzy_type() == fs.FUZZY_SETS.t1:
                membership = np.zeros((x.shape[0], len(rule_antecedents)))
            elif self.fuzzy_type() == fs.FUZZY_SETS.t2:
                membership = np.zeros((x.shape[0], len(rule_antecedents), 2))
            elif self.fuzzy_type() == fs.FUZZY_SETS.gt2:
                membership = np.zeros(
                    (x.shape[0], len(rule_antecedents), len(self.alpha_cuts), 2))

            n_nonvl = 0
            for ix, vl in enumerate(rule_antecedents):
                if vl >= 0:
                    membership_antecedent = list(antecedents_memberships[ix])[vl]
                    membership[:, ix] = membership_antecedent
                    n_nonvl += 1
                else:
                    membership[:, ix] = 1.0

            if n_nonvl == 0:
                membership[:, ix] = 0.0

            membership = self.tnorm(membership, axis=1)
            try:
                res[:, jx] = membership
            except UnboundLocalError:
                pass  # All the antecedents are dont care.

        if scaled:
            if self.fuzzy_type() == fs.FUZZY_SETS.t1:
                non_zero_rows = np.sum(res, axis=1) > 0
                res[non_zero_rows] = res[non_zero_rows] / \
                    np.sum(res[non_zero_rows], axis=1, keepdims=True)

            elif self.fuzzy_type() == fs.FUZZY_SETS.t2:
                non_zero_rows = np.sum(res[:, :, 0], axis=1) > 0
                res[non_zero_rows, :, 0] = res[non_zero_rows, :, 0] / \
                    np.sum(res[non_zero_rows, :, 0], axis=1, keepdims=True)
                non_zero_rows = np.sum(res[:, :, 1], axis=1) > 0
                res[non_zero_rows, :, 1] = res[non_zero_rows, :, 1] / \
                    np.sum(res[non_zero_rows, :, 1], axis=1, keepdims=True)
            
            elif self.fuzzy_type() == fs.FUZZY_SETS.gt2:

                for ix, alpha in enumerate(self.alpha_cuts):
                    relevant_res = res[:, :, ix, :]

                    non_zero_rows = np.sum(relevant_res[:, :, 0], axis=1) > 0
                    relevant_res[non_zero_rows, :, 0] = relevant_res[non_zero_rows, :, 0] / \
                        np.sum(relevant_res[non_zero_rows, :, 0], axis=1, keepdims=True)
                    non_zero_rows = np.sum(relevant_res[:, :, 1], axis=1) > 0
                    relevant_res[non_zero_rows, :, 1] = relevant_res[non_zero_rows, :, 1] / \
                        np.sum(relevant_res[non_zero_rows, :, 1], axis=1, keepdims=True)

                    res[:, :, ix, :] = relevant_res
                    
        return res


    def print_rules(self, return_rules:bool=False) -> None:
        '''
        Print the rules from the rule base.
        '''
        all_rules = ''
        for ix, rule in enumerate(self.rules):
            initiated = False
            str_rule = 'IF '
            for jx, vl_ix in enumerate(rule.antecedents):
                antecedent = self.antecedents[jx]
                keys = antecedent.linguistic_variable_names()

                if rule[jx] != -1:
                    if not initiated:
                        str_rule += str(antecedent.name) + ' IS ' + str(keys[vl_ix])
                        initiated = True
                    else:
                        str_rule += ' AND ' + str(antecedent.name) + \
                            ' IS ' + str(keys[vl_ix])

            try:
                score = rule.score if self.fuzzy_type() == fs.FUZZY_SETS.t1 else np.mean(rule.score)
                str_rule += ' WITH DS ' + str(score)

                # If the classification scores have been computed, print them.
                try:
                    str_rule += ', ACC ' + str(rule.accuracy)

                except AttributeError:
                    pass
            except AttributeError:
                str_rule += ' THEN consequent vl is ' + str(rule.consequent)
            
            all_rules += str_rule + '\n'

        if not return_rules:
            print(all_rules)
        else:
            return all_rules


    @abc.abstractmethod
    def inference(self, x: np.array) -> np.array:
        '''
        Computes the fuzzy output of the fl inference system.

        Return an array in shape samples x 2 (last is iv dimension)

        :param x: array with the values of the inputs.
        :return: array with the memberships of the consequents for each rule.
        '''
        raise NotImplementedError


    @abc.abstractmethod
    def forward(self, x: np.array) -> np.array:
        '''
        Computes the deffuzified output of the fl inference system.

        Return a vector of size (samples, )

        :param x: array with the values of the inputs.
        :return: array with the deffuzified output.
        '''
        raise NotImplementedError


    @abc.abstractmethod
    def fuzzy_type(self) -> fs.FUZZY_SETS:
        '''
        Returns the corresponding type of the RuleBase using the enum type in the fuzzy_sets module.

        :return: the type of fuzzy set used in the RuleBase.
        '''
        raise NotImplementedError


    def __len__(self):
        '''
        Returns the number of rules in the rule base.
        '''
        return len(self.rules)


    def prune_bad_rules(self, tolerance=0.01) -> None:
        '''
        Delete the rules from the rule base that do not have a dominance score superior to the threshold or have 0 accuracy in the training set.

        :param tolerance: threshold for the dominance score.
        '''
        delete_list = []
        try:
            for ix, rule in enumerate(self.rules):
                score = rule.score
                if (self.fuzzy_type() == fs.FUZZY_SETS.t2) or (self.fuzzy_type() == fs.FUZZY_SETS.gt2):
                    score = np.mean(score)

                try:
                    if score < tolerance or rule.accuracy == 0.0:
                        delete_list.append(ix)
                except AttributeError:
                    if score < tolerance:
                        delete_list.append(ix)
        except AttributeError:
            assert False, 'Dominance scores not computed for this rulebase'

        self.remove_rules(delete_list)


    def scores(self) -> np.array:
        '''
        Returns the dominance score for each rule.

        :return: array with the dominance score for each rule.
        '''
        scores = []
        for rule in self.rules:
            scores.append(rule.score)

        return np.array(scores)


    def __getitem__(self, item: int) -> RuleSimple:
        '''
        Returns the corresponding rulebase.

        :param item: index of the rule.
        :return: the corresponding rule.
        '''
        return self.rules[item]


    def __setitem__(self, key: int, value: RuleSimple) -> None:
        '''
        Set the corresponding rule.

        :param key: index of the rule.
        :param value: new rule.
        '''
        self.rules[key] = value
    

    def __iter__(self):
        '''
        Returns an iterator for the rule base.
        '''
        return iter(self.rules)


    def __str__(self):
        '''
        Returns a string with the rules in the rule base.
        '''
        str_res = ''
        for rule in self.rules:
            str_res += str(rule) + '\n'
        return str_res
    

    def __eq__(self, other):
        '''
        Returns True if the two rule bases are equal.
        '''
        return self.rules == other.rules
    

    def __hash__(self):
        '''
        Returns the hash of the rule base.
        '''
        return hash(str(self))
    

    def __add__(self, other):
        '''
        Adds two rule bases.
        '''
        return RuleBase(self.antecedents, self.rules + other.rules, self.consequent, self.tnorm)
    

    def n_linguist_variables(self) -> int:
        '''
        Returns the number of linguistic variables in the rule base.
        '''
        return len(self.antecedents)

    
class RuleBaseT2(RuleBase):
    '''
    Class optimized to work with multiple rules at the same time. Supports only one consequent. 
    (Use one rulebase per consequent to study classification problems. Check MasterRuleBase class for more documentation)

    This class supports iv approximation for t2 fs.
    '''

    def __init__(self, antecedents: list[fs.fuzzyVariable], rules: list[RuleSimple], consequent: fs.fuzzyVariable = None, tnorm=np.prod) -> None:
        '''
        Constructor of the RuleBaseT2 class.

        :param antecedents: list of fuzzy variables that are the antecedents of the rules.
        :param rules: list of rules.
        :param consequent: fuzzy variable that is the consequent of the rules.
        :param tnorm: t-norm used to compute the fuzzy output.
        '''
        rules = self.delete_rule_duplicates(rules)
        self.rules = rules
        self.antecedents = antecedents
        self.consequent = consequent
        self.tnorm = tnorm

        if consequent is not None:
            self.consequent_centroids = np.zeros(
                (len(consequent.linguistic_variable_names()), 2))

            for ix, vl_consequent in enumerate(consequent.linguistic_variables):
                consequent_domain = vl_consequent.domain
                domain_linspace = np.arange(
                    consequent_domain[0], consequent_domain[1], 0.05)
                consequent_memberships = vl_consequent.membership(
                    domain_linspace)

                self.consequent_centroids[ix, :] = centroid.compute_centroid_iv(
                    domain_linspace, consequent_memberships)

            self.consequent_centroids_rules = np.zeros((len(self.rules), 2))
            # If 0, we are classifying and we do not need the consequent centroids.
            if len(self.consequent_centroids) > 0:
                for ix, rule in enumerate(self.rules):
                    consequent_ix = rule.consequent
                    self.consequent_centroids_rules[ix] = self.consequent_centroids[consequent_ix]


    def inference(self, x: np.array) -> np.array:
        '''
        Computes the iv output of the t2 inference system.

        Return an array in shape samples x 2 (last is iv dimension)

        :param x: array with the values of the inputs.
        :return: array with the memberships of the consequents for each sample.
        '''
        res = np.zeros((x.shape[0], 2))

        antecedent_memberships = self.compute_rule_antecedent_memberships(x)
        for sample in range(antecedent_memberships.shape[0]):
            res[sample, :] = centroid.consequent_centroid(
                antecedent_memberships[sample], self.consequent_centroids_rules)

        return res


    def forward(self, x: np.array) -> np.array:
        '''
        Computes the deffuzified output of the t2 inference system.

        Return a vector of size (samples, )

        :param x: array with the values of the inputs.
        :return: array with the deffuzified output for each sample.
        '''
        return np.mean(self.inference(x))


    def fuzzy_type(self) -> fs.FUZZY_SETS:
        '''
        Returns the correspoing type of the RuleBase using the enum type in the fuzzy_sets module.

        :return: the corresponding fuzzy set type of the RuleBase.
        '''
        return fs.FUZZY_SETS.t2


class RuleBaseGT2(RuleBase):
    '''
    Class optimized to work with multiple rules at the same time. Supports only one consequent. 
    (Use one rulebase per consequent to study classification problems. Check MasterRuleBase class for more documentation)

    This class supports gt2 fs. (ONLY FOR CLASSIFICATION PROBLEMS)
    '''

    def __init__(self, antecedents: list[fs.fuzzyVariable], rules: list[RuleSimple], consequent: fs.fuzzyVariable = None, tnorm=np.prod) -> None:
        '''
        Constructor of the RuleBaseGT2 class.

        :param antecedents: list of fuzzy variables that are the antecedents of the rules.
        :param rules: list of rules.
        :param consequent: fuzzy variable that is the consequent of the rules.
        :param tnorm: t-norm used to compute the fuzzy output.
        '''
        rules = self.delete_rule_duplicates(rules)
        self.rules = rules
        self.antecedents = antecedents
        self.consequent = consequent
        self.tnorm = tnorm
        self.alpha_cuts = antecedents[0][0].alpha_cuts


    def inference(self, x: np.array) -> np.array:
        '''
        Computes the output of the gt2 inference system.

        Return an array in shape samples x alpha_cuts

        :param x: array with the values of the inputs.
        :return: array with the memberships of the consequents for each sample.
        '''
        res = np.zeros((x.shape[0], 2))

        antecedent_memberships = self.compute_rule_antecedent_memberships(x)
        for sample in range(antecedent_memberships.shape[0]):
            res[sample, :] = centroid.consequent_centroid(
                antecedent_memberships[sample], self.consequent_centroids_rules)

        return res


    def _alpha_reduction(self, x) -> np.array:
        '''
        Computes the type reduction to reduce the alpha cuts to one value.

        :param x: array with the values of the inputs.
        :return: array with the memberships of the consequents for each sample.
        '''
        formtatted = np.expand_dims(np.expand_dims(np.expand_dims(
            np.array(self.alpha_cuts), axis=1), axis=0), axis=0)
        return np.sum(formtatted * x, axis=2) / np.sum(self.alpha_cuts)


    def forward(self, x: np.array) -> np.array:
        '''
        Computes the deffuzified output of the t2 inference system.

        Return a vector of size (samples, )

        :param x: array with the values of the inputs.
        :return: array with the deffuzified output for each sample.
        '''
        return np.sum(np.array(self.alpha_cuts) * (self.inference(x)), axis=1) / np.sum(self.alpha_cuts)


    def fuzzy_type(self) -> fs.FUZZY_SETS:
        '''
        Returns the correspoing type of the RuleBase using the enum type in the fuzzy_sets module.

        :return: the corresponding fuzzy set type of the RuleBase.
        '''
        return fs.FUZZY_SETS.gt2


    def compute_rule_antecedent_memberships(self, x: np.array, scaled=True, antecedents_memberships=None) -> np.array:
        '''
        Computes the membership for the antecedents performing the alpha_cut reduction.

        :param x: array with the values of the inputs.
        :param scaled: if True, the memberships are scaled to sum 1 in each sample.
        :param antecedents_memberships: precomputed antecedent memberships. Not supported for GT2.
        :return: array with the memberships of the antecedents for each sample.
        '''
        antecedent_membership = super().compute_rule_antecedent_memberships(x, scaled)
        return self._alpha_reduction(antecedent_membership)


    def alpha_compute_rule_antecedent_memberships(self, x: np.array, scaled=True, antecedents_memberships=None) -> np.array:
        '''
        Computes the membership for the antecedents for all the alpha cuts.

        :param x: array with the values of the inputs.
        :param scaled: if True, the memberships are scaled to sum 1 in each sample.
        :param antecedents_memberships: precomputed antecedent memberships. Not supported for GT2.
        :return: array with the memberships of the antecedents for each sample.
        '''
        return super().compute_rule_antecedent_memberships(x, scaled)



class RuleBaseT1(RuleBase):
    '''
    Class optimized to work with multiple rules at the same time. Supports only one consequent.
    (Use one rulebase per consequent to study classification problems. Check MasterRuleBase class for more documentation)

    This class supports t1 fs.
    '''

    def __init__(self, antecedents: list[fs.fuzzyVariable], rules: list[RuleSimple], consequent: fs.fuzzyVariable = None, tnorm=np.prod) -> None:
        '''
        Constructor of the RuleBaseT1 class.

        :param antecedents: list of fuzzy variables that are the antecedents of the rules.
        :param rules: list of rules.
        :param consequent: fuzzy variable that is the consequent of the rules. ONLY on regression problems.
        :param tnorm: t-norm used to compute the fuzzy output.
        '''
        rules = self.delete_rule_duplicates(rules)
        self.rules = rules
        self.antecedents = antecedents
        self.consequent = consequent
        self.tnorm = tnorm

        if consequent is not None:
            self.consequent_centroids = np.zeros(
                (len(consequent.linguistic_variable_names()), ))

            for ix, vl_consequent in enumerate(consequent.linguistic_variables):
                consequent_domain = vl_consequent.domain
                domain_linspace = np.arange(
                    consequent_domain[0], consequent_domain[1], 0.05)
                consequent_memberships = vl_consequent.membership(
                    domain_linspace)

                self.consequent_centroids[ix] = centroid.center_of_masses(
                    domain_linspace, consequent_memberships)

            self.consequent_centroids_rules = np.zeros((len(self.rules), ))
            for ix, rule in enumerate(self.rules):
                consequent_ix = rule.consequent
                self.consequent_centroids_rules[ix] = self.consequent_centroids[consequent_ix]


    def inference(self, x: np.array) -> np.array:
        '''
        Computes the output of the t1 inference system.

        Return an array in shape samples.

        :param x: array with the values of the inputs.
        :return: array with the output of the inference system for each sample.
        '''
        res = np.zeros((x.shape[0]))

        antecedent_memberships = self.compute_rule_antecedent_memberships(x)

        for sample in range(antecedent_memberships.shape[0]):
            res[sample] = centroid.center_of_masses(
                self.consequent_centroids_rules, antecedent_memberships[sample])

        return res

    def forward(self, x: np.array) -> np.array:
        '''
        Same as inference() in the t1 case.

        Return a vector of size (samples, )

        :param x: array with the values of the inputs.
        :return: array with the deffuzified output for each sample.
        '''
        return self.inference(x)


    def fuzzy_type(self) -> fs.FUZZY_SETS:
        '''
        Returns the correspoing type of the RuleBase using the enum type in the fuzzy_sets module.

        :return: the corresponding fuzzy set type of the RuleBase.
        '''
        return fs.FUZZY_SETS.t1



class MasterRuleBase():
    '''
    This Class encompasses a list of rule bases where each one corresponds to a different class.
    '''

    def __init__(self, rule_base: list[RuleBase], consequent_names: list[str]=None) -> None:
        '''
        Constructor of the MasterRuleBase class.

        :param rule_base: list of rule bases.
        '''
        if len(rule_base) == 0:
            raise RuleError('No rule bases given!')
        
        self.rule_bases = rule_base
        self.antecedents = rule_base[0].antecedents

        if consequent_names is None:
            self.consequent_names = [ix for ix in range(len(self.rule_bases))]
        else:
            self.consequent_names = consequent_names


    def rename_cons(self, consequent_names: list[str]) -> None:
        '''
        Renames the consequents of the rule base.
        '''
        self.consequent_names = consequent_names


    def add_rule(self, rule: RuleSimple, consequent: int) -> None:
        '''
        Adds a rule to the rule base of the given consequent.

        :param rule: rule to add.
        :param consequent: index of the rule base to add the rule.
        '''
        self.rule_bases[consequent].add_rule(rule)
    

    def get_consequents(self) -> list[int]:
        '''
        Returns a list with the consequents of each rule base.

        :return: list with the consequents of each rule base.
        '''
        return sum([[ix]*len(x) for ix, x in enumerate(self.rule_bases)], [])


    def get_consequents_names(self) -> list[str]:
        '''
        Returns a list with the names of the consequents.

        :return: list with the names of the consequents.
        '''
        return self.consequent_names
    

    def get_rulebase_matrix(self) -> list[np.array]:
        '''
        Returns a list with the rulebases for each antecedent in matrix format.

        :return: list with the rulebases for each antecedent in matrix format.
        '''
        return [x.get_rulebase_matrix() for x in self.rule_bases]


    def get_scores(self) -> np.array:
        '''
        Returns the dominance score for each rule in all the rulebases.

        :return: array with the dominance score for each rule in all the rulebases.
        '''
        res = []
        for rb in self.rule_bases:
            res.append(rb.scores())

        res = [x for x in res if len(x) > 0]
        
        return np.concatenate(res, axis=0)


    def compute_firing_strenghts(self, X, precomputed_truth=None) -> np.array:
        '''
        Computes the firing strength of each rule for each sample.

        :param X: array with the values of the inputs.
        :param precomputed_truth: if not None, the antecedent memberships are already computed. (Used for sped up in genetic algorithms)
        :return: array with the firing strength of each rule for each sample.
        '''
        aux = []
        for ix in range(len(self.rule_bases)):
            aux.append(self[ix].compute_rule_antecedent_memberships(X, antecedents_memberships=precomputed_truth))

        # Firing strengths shape: samples x rules
        return np.concatenate(aux, axis=1)


    def _winning_rules(self, X: np.array, precomputed_truth=None) -> np.array:
        '''
        Returns the winning rule for each sample. Takes into account dominance scores if already computed.

        :param X: array with the values of the inputs.
        :return: array with the winning rule for each sample.
        '''
        
        firing_strengths = self.compute_firing_strenghts(X, precomputed_truth=precomputed_truth)

        association_degrees = self.get_scores() * firing_strengths

        if (self[0].fuzzy_type() == fs.FUZZY_SETS.t2) or (self[0].fuzzy_type() == fs.FUZZY_SETS.gt2):
            association_degrees = np.mean(association_degrees, axis=2)
        elif self[0].fuzzy_type() == fs.FUZZY_SETS.gt2:
            association_degrees = np.mean(association_degrees, axis=3)

        winning_rules = np.argmax(association_degrees, axis=1)

        return winning_rules
    
    
    def winning_rule_predict(self, X: np.array, precomputed_truth=None) -> np.array:
        '''
        Returns the winning rule for each sample. Takes into account dominance scores if already computed.

        :param X: array with the values of the inputs.
        :param precomputed_truth: if not None, the antecedent memberships are already computed. (Used for sped up in genetic algorithms)
        :return: array with the winning rule for each sample.
        '''
        # Raise an error if there no rules
        if len(self.get_rules()) == 0:
            raise RuleError('No rules to predict!')
        
        consequents = sum([[ix]*len(self[ix].rules)
                          for ix in range(len(self.rule_bases))], [])  # The sum is for flatenning the list
        winning_rules = self._winning_rules(X, precomputed_truth=precomputed_truth)

        return np.array([consequents[ix] for ix in winning_rules])


    def add_rule_base(self, rule_base: RuleBase) -> None:
        '''
        Adds a rule base to the list of rule bases.

        :param rule_base: rule base to add.
        '''
        self.rule_bases.append(rule_base)

        if len(self.rule_bases) != len(self.consequent_names):
            # We did not give proper names to the consequents
            self.consequent_names = [ix for ix in range(len(self.rule_bases))]


    def print_rules(self, return_rules=False) -> None:
        '''
        Print all the rules for all the consequents.

        :param autoprint: if True, the rules are printed. If False, the rules are returned as a string.
        '''
        res = ''
        for ix, ruleBase in enumerate(self.rule_bases):    
            res += 'Rules for consequent: ' + str(self.consequent_names[ix]) + '\n'
            
            res += '----------------\n'
            res += ruleBase.print_rules(return_rules=True) + '\n'
        
        
        if return_rules:
            return res
        else:
            print(res)


    def get_rules(self) -> list[RuleSimple]:
        '''
        Returns a list with all the rules.

        :return: list with all the rules.
        '''
        return [rule for ruleBase in self.rule_bases for rule in ruleBase.rules]


    def fuzzy_type(self) -> fs.FUZZY_SETS:
        '''
        Returns the correspoing type of the RuleBase using the enum type in the fuzzy_sets module.

        :return: the corresponding fuzzy set type of the RuleBase.
        '''
        return self.rule_bases[0].fuzzy_type()


    def purge_rules(self, tolerance=0.001) -> None:
        '''
        Delete the roles with a dominance score lower than the tolerance.

        :param tolerance: tolerance to delete the rules.
        '''
        for ruleBase in self.rule_bases:
            ruleBase.prune_bad_rules(tolerance)


    def __getitem__(self, item) -> RuleBase:
        '''
        Returns the corresponding rulebase.

        :param item: index of the rulebase.
        :return: the corresponding rulebase.
        '''
        return self.rule_bases[item]


    def __len__(self) -> int:
        '''
        Returns the number of rule bases.
        '''
        return len(self.rule_bases)


    def __str__(self) -> str:
        '''
        Returns a string with the rules for each consequent.
        '''
        return self.print_rules(return_rules=True)
    

    def __eq__(self, __value: object) -> bool:
        '''
        Returns True if the two rule bases are equal.

        :param __value: object to compare.
        :return: True if the two rule bases are equal.
        '''
        if not isinstance(__value, MasterRuleBase):
            return False
        else:
            return self.rule_bases == __value.rule_bases
        
    
    def __call__(self, X: np.array) -> np.array:
        '''
        Gives the prediction for each sample (same as winning_rule_predict)

        :param X: array of dims: samples x features.
        :return: vector of predictions, size: samples,
        '''
        return self.winning_rule_predict(X)
        
        
    def get_rulebases(self) -> list[RuleBase]:
        '''
        Returns a list with all the rules.

        :return: list
        '''
        return self.rule_bases


    def n_linguist_variables(self) -> int:
        '''
        Returns the number of linguistic variables in the rule base.
        '''
        return len(self.antecedents[0])
    

    def get_antecedents(self) -> list[fs.fuzzyVariable]:
        '''
        Returns the antecedents of the rule base.
        '''
        return self.antecedents
    

def construct_rule_base(rule_matrix: np.array, consequents: np.array, antecedents: list[fs.fuzzyVariable], rule_weights: np.array, class_names: list=None) -> MasterRuleBase:
    '''
    Constructs a rule base from a matrix of rules.

    :param rule_matrix: matrix with the rules.
    :param consequents: array with the consequents per rule.
    :param antecedents: list of fuzzy variables.
    :param class_names: list with the names of the classes.
    '''
    rule_lists = {ix:[] for ix in range(len(np.unique(consequents)))}
    fs_studied = antecedents[0].fuzzy_type()
    for ix, consequent in enumerate(consequents):
        if not np.equal(rule_matrix[ix], -1).all():
            rule_object = RuleSimple(rule_matrix[ix])
            rule_object.score = rule_weights[ix]
            rule_lists[consequent].append(rule_object)

    for ix, consequent in enumerate(np.unique(consequents)):
        if fs_studied == fs.FUZZY_SETS.t1:
            rule_base = RuleBaseT1(antecedents, rule_lists[ix])
        elif fs_studied == fs.FUZZY_SETS.t2:
            rule_base = RuleBaseT2(antecedents, rule_lists[ix])
        elif fs_studied == fs.FUZZY_SETS.gt2:
            rule_base = RuleBaseGT2(antecedents, rule_lists[ix])
        
        if ix == 0:
            res = MasterRuleBase([rule_base], np.unique(consequents))
        else:
            res.add_rule_base(rule_base)

    if class_names is not None:
        res.rename_cons(class_names)

    return res


def list_rules_to_matrix(rule_list: list[RuleSimple]) -> np.array:
    '''
    Returns a matrix out of the rule list.

    :param rule_list: list of rules.
    :return: matrix with the antecedents of the rules.
    '''
    assert len(rule_list) > 0, 'No rules to list!'

    res = np.zeros((len(rule_list), len(rule_list[0].antecedents)))
    for ix, rule in enumerate(rule_list):
        res[ix, :] = rule.antecedents

    return res




import numpy as np
import math

import sys
sys.path.append('./ex_fuzzy/')
sys.path.append('../ex_fuzzy/')
import ex_fuzzy as ex_fuzzy


def test_fuzzy_t1_memberships():
    '''
    Tests that fuzzy t1 memberships compute correctly.
    '''
    trial_fs = ex_fuzzy.fuzzy_sets.FS('trial', [0,0.1, 0.25, 0.5], [0,1])

    assert trial_fs.name == 'trial', 'Name not correctly set'
    assert trial_fs(0.1) == 1, 'Trapezoidal membership fails in first term'
    assert trial_fs(1) == 0, 'Trapezoidal membership fails in last term'
    assert trial_fs(0.51) == 0, 'Trapezoidal memberships does not stop where it shoud'
    assert trial_fs(0.05) == 0.5, 'Trapeziodal membership increasining in non linear way' 


def test_fuzzy_t2_memberships():
    '''
    Tests that fuzzy t2 (iv) memberships compute correctly.
    '''
    trial_fs = ex_fuzzy.fuzzy_sets.IVFS('trial', [0,0.1, 0.25, 0.4], [0, 0.15, 0.25, 0.5], [0,1], lower_height=0.8)

    assert trial_fs.name == 'trial', 'Name not correctly set'
    trial_mfs = trial_fs(0.2)
    assert trial_mfs[0] <= trial_mfs[1], 'Incoherent lower and upper membership behaviour'


def test_fuzzy_gt2_memberships():
    '''
    Tests that fuzzy gt2 memberships compute correctly.
    '''
    trial_fs = {}
    for x in [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 1]:
        trial_fs[x] = ex_fuzzy.fuzzy_sets.FS('trial', [0.25,0.5, 0.5, 0.75], [0,1])

    trial_fs = ex_fuzzy.fuzzy_sets.GT2('trial', trial_fs, [0, 1], significant_decimals=1, unit_resolution=0.001)
    res = trial_fs(0.3)
    assert math.isclose(np.mean(trial_fs.alpha_reduction(res)), 0.5, abs_tol=0.1), 'GT2 memberships not correctly reduced'



    
if __name__ == '__main__':
    test_fuzzy_t1_memberships()
    test_fuzzy_t2_memberships()
    test_fuzzy_gt2_memberships()



import numpy as np
import pandas as pd
import math
import sys
sys.path.append('../ex_fuzzy/')
sys.path.append('./ex_fuzzy/')

import ex_fuzzy

sample_size = 1000

def test_random_classification(fs_type=ex_fuzzy.fuzzy_sets.FUZZY_SETS.t1):
    global sample_size

    sample = np.random.random_sample((sample_size, 5))
    targets = np.random.randint(0, 2, sample_size)

    model = ex_fuzzy.evolutionary_fit.BaseFuzzyRulesClassifier(10, 3, fs_type, verbose=False, tolerance=0.0, n_linguist_variables=3)
    model.fit(sample, targets)
    predictions = model.predict(sample)
    assert math.isclose(np.mean(np.equal(predictions, targets)), 0.5, abs_tol=0.1)


def test_random_classification_precomputed(fs_type=ex_fuzzy.fuzzy_sets.FUZZY_SETS.t1):
    global sample_size

    sample = np.random.random_sample((sample_size, 5))
    targets = np.random.randint(0, 2, sample_size)
    vl_partitions = ex_fuzzy.utils.construct_partitions(sample, fs_type)
    model = ex_fuzzy.evolutionary_fit.BaseFuzzyRulesClassifier(10, 3, verbose=False, tolerance=0.0, linguistic_variables=vl_partitions)
    model.fit(sample, targets)
    predictions = model.predict(sample)
    assert math.isclose(np.mean(np.equal(predictions, targets)), 0.5, abs_tol=0.1)


def test_random_classification_t2():
    test_random_classification(ex_fuzzy.fuzzy_sets.FUZZY_SETS.t2)


def test_random_classification_t2_precomputed():
    test_random_classification_precomputed(ex_fuzzy.fuzzy_sets.FUZZY_SETS.t2)


def test_random_classification_gt2():
    test_random_classification(ex_fuzzy.fuzzy_sets.FUZZY_SETS.t2)


def test_random_classification_gt2_precomputed():
    test_random_classification_precomputed(ex_fuzzy.fuzzy_sets.FUZZY_SETS.t2)

if __name__ == '__main__':
    test_random_classification_t2()
    test_random_classification_t2_precomputed()
    test_random_classification_gt2()
    test_random_classification_gt2_precomputed()



# -*- coding: utf-8 -*-
"""
This is a the source file that contains the functions necessary to visualize the set of rules.
"""
import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import networkx as nx

try:
    from . import rules
    from . import evolutionary_fit as evf
    from . import fuzzy_sets as fs
    from . import maintenance as mnt
except ImportError:
    import rules
    import evolutionary_fit as evf
    import fuzzy_sets as fs
    import maintenance as mnt


def _column_histogram(rule_matrix: np.array) -> dict:
    '''
    Computes the histogram for all the unique values in a column.

    :param rule_matrix: vector with the numerical values.
    :return: dictionary with the histogram.
    '''
    res = {}
    for x in np.unique(rule_matrix):
        if x != -1:
            res[x] = np.sum(rule_matrix == x)

    return res


def _histogram(rule_matrix: np.array) -> list[dict]:
    '''
    Returns a list with the histogram for each antecedent according to linguist variables.

    :param rule_matrix: matrix with the rules.
    :return: list with the histogram in dictionary format for each antecedent.
    '''
    res = []

    for column_ix in range(rule_matrix.shape[1]):
        res.append(_column_histogram(rule_matrix[:, column_ix]))

    return res


def _max_values(appearances: list) -> tuple[int, int]:
    '''
    Returns the antecedent and its linguistic value most repeated.

    :param appearances: list with the histogram for each antecedent.
    :return: tuple with the antecedent and its linguistic value most repeated.
    '''
    res = 0
    antecedent = None
    vl = None

    for ix, apperances_an in enumerate(appearances):
        for key, value in apperances_an.items():
            if value > res:
                res = value
                antecedent = ix
                vl = key

    return antecedent, vl


def create_graph_connection(rules, possible_vl):
    '''
    Returns a square matrix where each number indicates if two nodes are connected.
    Connectes by checking if both are in the same rule.

    :param rules: list with the rules.
    :param possible_vl: number of linguistic variables.
    :return: square matrix where each number indicates if two nodes are connected.
    '''
    def generate_index(ant, vl0): return int(possible_vl * ant + vl0)
    res = np.zeros(
        (possible_vl * rules.shape[1], possible_vl * rules.shape[1]))

    for rule in rules:
        for antecedent, vl in enumerate(rule):
            if vl > -1:
                for antecedent2, vl2 in enumerate(rule):
                    if vl2 > -1:
                        res_index1 = generate_index(antecedent, vl)
                        res_index2 = generate_index(antecedent2, vl2)

                        res[res_index1, res_index2] += 1

    return res / 2


def choose_popular_rules(rule_matrix: np.array) -> np.array:
    '''
    Returns the index of the rules that contain the most popular antecedent in the dataset.

    :param rule_matrix: matrix with the rules.
    :return: numpy array with the rules that contain the most popular antecedent in the dataset.
    '''
    appearances = _histogram(rule_matrix)
    antecedent, vl = _max_values(appearances)

    chosen_rules = rule_matrix[:, antecedent] == vl

    return chosen_rules


def connect_rulebase(rulebase: rules.RuleBase) -> list[np.array]:
    '''
    Connects antecedents connected by checking if both are in the same rule.

    :param rulebase: Rule base to connect.
    :return: List of pandas dataframes with the connections in adjacency matrix format.
    '''

    # We choose those rules to explain, as those who have the most common antecedent.
    rule_matrix = rules.list_rules_to_matrix(rulebase.rules)
    res = []
    antecedents_names = [
        x.name + ' ' + y for x in rulebase.antecedents for y in rulebase.antecedents[0].linguistic_variable_names()]

    try:
        while rule_matrix.shape[0] > 0:
            rules_to_viz = rule_matrix[choose_popular_rules(rule_matrix), :]
            rule_matrix = rule_matrix[(
                1 - choose_popular_rules(rule_matrix)).astype(bool), :]

            graph_rule = create_graph_connection(rules_to_viz, len(
                rulebase.antecedents[0].linguistic_variable_names()))
            res.append(pd.DataFrame(
                graph_rule, columns=antecedents_names, index=antecedents_names))
    except:
        print('Error in the visualization of the rule base, probably because the rulebase is too small')

    return res


def connect_master_rulebase(mrule_base: rules.MasterRuleBase) -> list[list[np.array]]:
    '''
    Connects antecedents connected by checking if both are in the same rule.

    :param mrule_base: Master rule base to connect.
    :return: List of list of pandas dataframes with the connections in adjacency matrix format.
    '''
    res = []
    for ix, rule_base in enumerate(mrule_base.rule_bases):
        try:
            res.append(connect_rulebase(rule_base))
        except:
            print('Error in the visualization of the rule base: "' + str(mrule_base.get_consequents_names()[ix]) + '", probably because there are no rules in the rule base.')

    return res


def visualize_rulebase(mrule_base: rules.MasterRuleBase, export_path: str=None) -> None:
    '''
    Visualize a rule base using low, medium and high partitions with 1 rule in common -> 1 edge connections.

    :param mrule_base: Master rule base to visualize.
    :param export_path: Path to export the graph.
    '''

    if mnt.save_usage_flag:
        mnt.usage_data[mnt.usage_categories.Visualization]['plot_graph'] += 1

    def color_func(a):
        '''
        Returns the color of the node according to the linguistic variable.

        :param a: Node name.
        :return: Color of the node.
        '''
        node_colors = ['blue', 'yellow', 'red']

        if ' L' in a:
            return node_colors[0]
        elif ' M' in a:
            return node_colors[1]
        else:
            return node_colors[2]

    def vl_prune(a): return a.replace('High', 'H').replace(
        'Medium', 'M').replace('Low', 'L').strip()

    if isinstance(mrule_base, evf.BaseFuzzyRulesClassifier):
        mrule_base = mrule_base.rule_base

    connected_mrule_base = connect_master_rulebase(mrule_base)

    for ix, connected_rule_base in enumerate(connected_mrule_base):
        for jx, rule in enumerate(connected_rule_base):
            G = nx.from_pandas_adjacency(rule)
            isolated_nodes = [
                node for node in G.nodes() if G.degree(node) == 0]
            G.remove_nodes_from(isolated_nodes)
            auto_edges = nx.selfloop_edges(G)
            G.remove_edges_from(auto_edges)

            new_node_names = [vl_prune(node) for node in G.nodes()]
            mapping = dict(zip(G, new_node_names))
            G = nx.relabel_nodes(G, mapping)

            if jx == 0:
                G_final = G
            else:
                G_final = nx.compose(G_final, G)

        fig, ax = plt.subplots()
        try:
            os = nx.nx_agraph.graphviz_layout(G_final, prog='sfdp')
        except ImportError:
            os = nx.kamada_kawai_layout(G_final)

        node_colors = [color_func(node) for node in G_final.nodes()]
        nx.draw(G_final, with_labels=True, ax=ax,
                pos=os, node_color=node_colors)
        plt.title('Consequent: ' + str(ix))
        fig.show()

        if export_path is not None:
            nx.write_gexf(G_final, os.path.join(export_path,
                          'consequent_' + str(ix) + '.gexf'))


def plot_fuzzy_variable(fuzzy_variable: fs.fuzzyVariable) -> None:
    '''
    Plots a fuzzy variable using trapezoidal membership functions.

    :param fuzzy_variable: a fuzzy variable from the fuzzyVariable class in fuzzy_set module.
    :return: None
    '''
    if mnt.save_usage_flag:
        mnt.usage_data[mnt.usage_categories.Visualization]['plot_fuzzy_variable'] += 1

    if fuzzy_variable.linguistic_variables[0].type() != fs.FUZZY_SETS.gt2:
        fig, ax = plt.subplots()
    else:
        fig = plt.figure()
        ax = plt.axes(projection='3d')

    memberships = [0, 1, 1, 0]

    colors = ['b', 'r', 'g', 'orange', 'y']
    for ix, fuzzy_set in enumerate(fuzzy_variable.linguistic_variables):
        name = fuzzy_set.name
        initiated = False
        fz_studied =  fuzzy_set.type()

        if  fz_studied == fs.FUZZY_SETS.t1:
            ax.plot(fuzzy_set.membership_parameters,
                    memberships, colors[ix], label=name)
        elif fz_studied == fs.FUZZY_SETS.t2:
            ax.plot(fuzzy_set.secondMF_lower, np.array(memberships) * fuzzy_set.lower_height, 'black')
            ax.plot(fuzzy_set.secondMF_upper, np.array(memberships), 'black')

            # Compute the memberships for the lower/upper membership points. We do it in this way because non-exact 0/1s give problems.
            x_lower = fuzzy_set.secondMF_lower
            x_lower_lmemberships = [0.0 ,fuzzy_set.lower_height ,fuzzy_set.lower_height, 0.0] 
            x_lower_umemberships = [fuzzy_set(x_lower[0])[1] , 1.0, 1.0 , fuzzy_set(x_lower[3])[1]]

            x_upper = fuzzy_set.secondMF_upper
            x_upper_lmemberships  = [0.0 , fuzzy_set(x_upper[1])[0], fuzzy_set(x_upper[2])[0], 0.0] 
            x_upper_umemberships  = [0.0 ,1.0 ,1.0, 0.0] 

            x_values = list(x_lower) + list(x_upper)
            lmembership_values = list(x_lower_lmemberships) + list(x_upper_lmemberships)
            umembership_values = list(x_lower_umemberships) + list(x_upper_umemberships)
            aux_df = pd.DataFrame(zip(x_values, lmembership_values, umembership_values),  columns=['x', 'l', 'u'])
            

            if len(aux_df['x']) != len(set(aux_df['x'])): # There are repeated elements, so we use an order that should work in this case
                # u0 l0 u1 l1 l2 u2 l3 u3
                x = list((x_upper[0], x_lower[0], x_upper[1], x_lower[1], x_lower[2], x_upper[2], x_lower[3], x_upper[3]))
                l_memberships = list((x_upper_lmemberships[0], x_lower_lmemberships[0], x_upper_lmemberships[1], x_lower_lmemberships[1], x_lower_lmemberships[2], x_upper_lmemberships[2], x_lower_lmemberships[3], x_upper_lmemberships[3]))
                u_memberships = list((x_upper_umemberships[0], x_lower_umemberships[0], x_upper_umemberships[1], x_lower_umemberships[1], x_lower_umemberships[2], x_upper_umemberships[2], x_lower_umemberships[3], x_upper_umemberships[3]))

                ax.fill_between(x, l_memberships, u_memberships, color=colors[ix], alpha=0.5, label=name)
            else:
                aux_df.sort_values('x', inplace=True)
                ax.fill_between(aux_df['x'], aux_df['l'], aux_df['u'], color=colors[ix], alpha=0.5, label=name)

        elif fz_studied == fs.FUZZY_SETS.gt2:
            for key, value in fuzzy_set.secondary_memberships.items():
                
                gt2_memberships = value(fuzzy_set.sample_unit_domain)
                key_plot = [float(key)]*sum(gt2_memberships > 0)
                if initiated:
                    ax.plot(key_plot, fuzzy_set.sample_unit_domain[gt2_memberships > 0], gt2_memberships[gt2_memberships > 0], color=colors[ix])
                else:
                    ax.plot(key_plot,  fuzzy_set.sample_unit_domain[gt2_memberships > 0], gt2_memberships[gt2_memberships > 0], color=colors[ix], label=name)
                    initiated = True

    ax.legend(loc='upper right', shadow=True)
    plt.title(fuzzy_variable.name)
    fig.show()


def matrix_rule_base_form(rule_base: rules.Rule) -> pd.DataFrame:
    '''
    Returns a matrix with the rule base in the form of a matrix to visualize.

    :param mrule_base: Rule base to transform.
    :return: Matrix with the rule base in the form of a matrix.
    '''

    n_rules = len(rule_base.rules)
    antecedents = len(rule_base.antecedents)

    res = pd.DataFrame(np.zeros((n_rules, antecedents)), columns=[jx.name for jx in rule_base.antecedents])

    for ix, rule in enumerate(rule_base):
        for jx, antecedent in enumerate(rule_base.antecedents):
            res.loc[ix, antecedent.name] = rule.antecedents[jx]
    
    return res


def filter_useless_columns(df: pd.DataFrame) -> pd.DataFrame:
    '''
    Filter columns with only one value.

    :param df: Dataframe to filter.
    :return: Filtered dataframe.
    '''
    for column in df.columns:
        if df[column].unique()[0] == -1:
            df.drop(column, axis=1, inplace=True)

    return df


def rules_to_latex(rule_base:rules.MasterRuleBase) -> str:
    '''
    Prints the rule base in a latex format. It prints

    :note: if the rule base has three different linguistic labels, it will use custom commands for the partitions. You can define these commands (\low, \mid, \hig, \dc) to show colors, figures, etc. Be sure to recheck the DS, ACC values in this case, because 1.0 values of them are also converted to these commands.
    
    :param rule_base: the master rule base to print.
    :returns: the String as a latex tabular.
    '''
    class proxy_dict():

        def __init__(self) -> None:
            self.cell_colors = {
                -1: '\\dc',
                0: '\\low',
                1: '\\med',
                2: '\\hig'
            }
        
        def __getitem__(self, value) -> str:
            if value in self.cell_colors.keys():
                return self.cell_colors[value]
            else:
                return "{:.2f}".format(value)


    # Define the mapping for cell colors
    rules_matrix = rule_base.get_rulebase_matrix()

    # Add the consequent to the rules
    cons_rules_matrix = []
    for ix_cons, ruls in enumerate(rules_matrix):
        for rule_list in ruls:
            if len(rule_list.shape) == 2:
                cons_rules_matrix.append(np.append(np.ones((rule_list.shape[0])) * ix_cons, rule_list, axis=1))
            else:
                cons_rules_matrix.append(np.append(np.ones((1, )) * ix_cons, rule_list))
                     
    dominance_scores = np.array([x.score for x in rule_base.get_rules()])
    if len(dominance_scores.shape) == 2:
        dominance_scores = np.mean(dominance_scores, axis=1, keepdims=True)
    else:
        dominance_scores = np.expand_dims(dominance_scores, axis=1)

    accs = np.array([x.accuracy for x in rule_base.get_rules()])
    accs = np.expand_dims(accs, axis=1)
    cons_rules_matrix = np.append(np.append(np.array(cons_rules_matrix), dominance_scores, axis=1), accs, axis=1)
    column_order = ['Consequent'] + [a.name for a in rule_base.antecedents] + ['DS', 'Acc']
    df = pd.DataFrame(cons_rules_matrix, columns=column_order)
    cell_colors = proxy_dict()

    # Create the LaTeX table
    latex_table = "\\begin{tabular}{" + "c" * (len(column_order)-2) + "|cc}\n"
    latex_table += "\t\\toprule\n"
    latex_table += "\t" + " & ".join(column_order) + " \\\\\n"
    latex_table += "\t\\midrule\n"

    i = 0
    for cluster, group in df.groupby('Consequent'):
            latex_table += f"\t\\multirow{{{len(group)}}}{{*}}{{{cluster}}}"
            for _, row in group.iterrows():
                if i % 2 == 0: # Add a shade of grey
                    latex_table += " & \cellcolor{gray!25}" + " & \cellcolor{gray!25}".join([cell_colors[val] for val in row[column_order[1:]]]) + " \\\\\n"
                else:
                    latex_table += " & " + " & ".join([cell_colors[val] for val in row[column_order[1:]]]) + " \\\\\n"
                i += 1
            if cluster != len(rules_matrix) - 1:
                latex_table += "\t\\midrule\n"            
                
    latex_table += "\t\\bottomrule\n"
    latex_table += "\\end{tabular}"

    print(latex_table)


import numpy as np
import math

import sys
sys.path.append('../ex_fuzzy/')

import ex_fuzzy as ex_fuzzy

def test_t2_centroid():
    '''
    Tests that fuzzy t2 (iv) centroids compute correctly.
    '''
    trial_fs = ex_fuzzy.fuzzy_sets.IVFS('trial', [0.35, 0.4, 0.6, 0.65], [0.25 ,0.4, 0.6, 0.75], [0,1], lower_height=0.8)
    z = np.arange(0, 1, 0.001)
    assert math.isclose(ex_fuzzy.centroid.compute_centroid_iv(z, trial_fs(z))[0], 0.5, abs_tol=0.01), 'T2 centroid right not correctly computed'
    assert math.isclose(ex_fuzzy.centroid.compute_centroid_iv(z, trial_fs(z))[1], 0.5, abs_tol=0.01), 'T2 centroid left not correctly computed'


def test_t2_centroid_centroids():
    '''
    Tests that fuzzy t2 (iv) centroid computed from previous centroids is computed correctly.
    '''
    trial_fs = ex_fuzzy.fuzzy_sets.IVFS('trial', [0.35, 0.4, 0.6, 0.65], [0.25 ,0.4, 0.6, 0.75], [0,1], lower_height=0.8)
    z = np.arange(0, 1, 0.001)
    assert math.isclose(ex_fuzzy.centroid.compute_centroid_t2_r(z, trial_fs(z)), 0.5, abs_tol=0.01), 'T2 centroid right not correctly computed'
    assert math.isclose(ex_fuzzy.centroid.compute_centroid_t2_l(z, trial_fs(z)), 0.5, abs_tol=0.01), 'T2 centroid left not correctly computed'





"""
This is a the source file that contains the class to train/fit the rulebase using a genetic algorithm.

"""
import numpy as np
import pandas as pd

from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import matthews_corrcoef
from sklearn.base import ClassifierMixin
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.core.problem import Problem
from pymoo.optimize import minimize
from pymoo.operators.sampling.rnd import IntegerRandomSampling
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PolynomialMutation
from pymoo.core.variable import Integer
from multiprocessing.pool import ThreadPool
from pymoo.core.problem import StarmapParallelization

try:
    from . import fuzzy_sets as fs
    from . import rules
    from . import eval_rules as evr
    from . import vis_rules
    from . import maintenance as mnt
except ImportError:
    import fuzzy_sets as fs
    import rules
    import eval_rules as evr
    import vis_rules
    import maintenance as mnt



class BaseFuzzyRulesClassifier(ClassifierMixin):
    '''
    Class that is used as a classifier for a fuzzy rule based system. Supports precomputed and optimization of the linguistic variables.
    '''

    def __init__(self,  nRules: int = 30, nAnts: int = 4, fuzzy_type: fs.FUZZY_SETS = fs.FUZZY_SETS.t1, tolerance: float = 0.0, class_names: list[str] = None,
                 n_linguist_variables: int = 3, verbose=False, linguistic_variables: list[fs.fuzzyVariable] = None,
                 domain: list[float] = None, n_class: int=None, precomputed_rules: rules.MasterRuleBase=None, runner: int=1) -> None:
        '''
        Inits the optimizer with the corresponding parameters.

        :param nRules: number of rules to optimize.
        :param nAnts: max number of antecedents to use.
        :param fuzzy type: FUZZY_SET enum type in fuzzy_sets module. The kind of fuzzy set used.
        :param tolerance: tolerance for the dominance score of the rules.
        :param n_linguist_variables: number of linguistic variables per antecedent.
        :param verbose: if True, prints the progress of the optimization.
        :param linguistic_variables: list of fuzzyVariables type. If None (default) the optimization process will init+optimize them.
        :param domain: list of the limits for each variable. If None (default) the classifier will compute them empirically.
        :param n_class: names of the classes in the problem. If None (default) the classifier will compute it empirically.
        :param precomputed_rules: MasterRuleBase object. If not None, the classifier will use the rules in the object and ignore the conflicting parameters.
        :param runner: number of threads to use. If None (default) the classifier will use 1 thread.
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.Funcs]['fit'] += 1

        if precomputed_rules is not None:
            self.nRules = len(precomputed_rules.get_rules())
            self.nAnts = len(precomputed_rules.get_rules()[0].antecedents)
            self.n_class = len(precomputed_rules)
            self.nclasses_ = len(precomputed_rules.consequent_names)
            self.classes_names = precomputed_rules.consequent_names
            self.rule_base = precomputed_rules
        else:
            self.nRules = nRules
            self.nAnts = nAnts
            self.nclasses_ = n_class
            if not (class_names is None):
                if isinstance(class_names, np.ndarray):
                    self.classes_names = list(class_names)
                else:
                    self.classes_names = class_names
            else:
                self.classes_names = class_names

        self.custom_loss = None
        self.verbose = verbose
        self.tolerance = tolerance
        

        if runner > 1:
            pool = ThreadPool(runner)
            self.thread_runner = StarmapParallelization(pool.starmap)
        else:
            self.thread_runner = None
        
        if linguistic_variables is not None:
            if mnt.save_usage_flag:
                mnt.usage_data[mnt.usage_categories.Funcs]['precompute_labels'] += 1
            # If the linguistic variables are precomputed then we act accordingly
            self.lvs = linguistic_variables
            self.n_linguist_variables = len(
                self.lvs[0].linguistic_variable_names())
            self.domain = None
            self.fuzzy_type = self.lvs[0].fuzzy_type()
        else:
            if mnt.save_usage_flag:
                mnt.usage_data[mnt.usage_categories.Funcs]['opt_labels'] += 1

            # If not, then we need the parameters sumistered by the user.
            self.lvs = None
            self.fuzzy_type = fuzzy_type
            self.n_linguist_variables = n_linguist_variables
            self.domain = domain

        self.alpha_ = 0.0
        self.beta_ = 0.0

    def customized_loss(self, loss_function):
        '''
        Function to customize the loss function used for the optimization.

        :param loss_function: function that takes as input the true labels and the predicted labels and returns a float.
        :return: None
        '''
        self.custom_loss = loss_function


    def fit(self, X: np.array, y: np.array, n_gen:int=70, pop_size:int=30,
            checkpoints:int=0, candidate_rules:rules.MasterRuleBase=None, initial_rules:rules.MasterRuleBase=None, random_state:int=33) -> None:
        '''
        Fits a fuzzy rule based classifier using a genetic algorithm to the given data.

        :param X: numpy array samples x features
        :param y: labels. integer array samples (x 1)
        :param n_gen: integer. Number of generations to run the genetic algorithm.
        :param pop_size: integer. Population size for each gneration.
        :param checkpoints: integer. Number of checkpoints to save the best rulebase found so far.
        :param candidate_rules: if these rules exist, the optimization process will choose the best rules from this set. If None (default) the rules will be generated from scratch.
        :return: None. The classifier is fitted to the data.
        '''
        if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.Funcs]['fit'] += 1
            
        if self.classes_names is None:
            self.classes_names = [aux for aux in np.unique(y)]
        
        if isinstance(y[0], str):
            y = np.array([self.classes_names.index(str(aux)) for aux in y])
            
        if candidate_rules is None:
            if initial_rules is not None:
                self.fuzzy_type = initial_rules.fuzzy_type()
                self.n_linguist_variables = initial_rules.n_linguist_variables()
                self.domain = [fv.domain for fv in initial_rules[0].antecedents]
                self.nRules = len(initial_rules.get_rules())
                self.nAnts = len(initial_rules.get_rules()[0].antecedents)

            if self.lvs is None:
                # If Fuzzy variables need to be optimized.
                problem = FitRuleBase(X, y, nRules=self.nRules, nAnts=self.nAnts, tolerance=self.tolerance, n_classes=len(np.unique(y)),
                                    n_linguist_variables=self.n_linguist_variables, fuzzy_type=self.fuzzy_type, domain=self.domain, thread_runner=self.thread_runner,
                                    alpha=self.alpha_, beta=self.beta_)
            else:
                # If Fuzzy variables are already precomputed.
                problem = FitRuleBase(X, y, nRules=self.nRules, nAnts=self.nAnts, n_classes=len(np.unique(y)),
                                    linguistic_variables=self.lvs, domain=self.domain, tolerance=self.tolerance, thread_runner=self.thread_runner,
                                    alpha=self.alpha_, beta=self.beta_)
        else:
            self.fuzzy_type = candidate_rules.fuzzy_type()
            self.n_linguist_variables = candidate_rules.n_linguist_variables()
            problem = ExploreRuleBases(X, y, n_classes=len(np.unique(y)), candidate_rules=candidate_rules, thread_runner=self.thread_runner, nRules=self.nRules)

        if self.custom_loss is not None:
            problem.fitness_func = self.custom_loss

        if initial_rules is not None:
            rules_gene = problem.encode_rulebase(initial_rules, self.lvs is None)
            rules_gene = (np.ones((pop_size, len(rules_gene))) * rules_gene).astype(int)
        else:
            rules_gene = IntegerRandomSampling()

        algorithm = GA(
            pop_size=pop_size,
            crossover=SBX(prob=.3, eta=3.0),
            mutation=PolynomialMutation(eta=7.0),
            sampling=rules_gene,
            eliminate_duplicates=False)
        

        if checkpoints > 0:
            if self.verbose:
                print('=================================================')
                print('n_gen  |  n_eval  |     f_avg     |     f_min    ')
                print('=================================================')
            algorithm.setup(problem, seed=33, termination=('n_gen', n_gen)) # 33? Soon...
            for k in range(n_gen):
                algorithm.next()
                res = algorithm
                if self.verbose:
                    print('%-6s | %-8s | %-8s | %-8s' % (res.n_gen, res.evaluator.n_eval, res.pop.get('F').mean(), res.pop.get('F').min()))
                if k % checkpoints == 0:
                    with open("checkpoint_" + str(algorithm.n_gen), "w") as f:
                        pop = algorithm.pop
                        fitness_last_gen = pop.get('F')
                        best_solution_arg = np.argmin(fitness_last_gen)
                        best_individual = pop.get('X')[best_solution_arg, :]

                        rule_base = problem._construct_ruleBase(
                            best_individual, self.fuzzy_type)
                        eval_performance = evr.evalRuleBase(
                            rule_base, np.array(X), y)
                        
                        eval_performance.add_full_evaluation()  
                        # self.rename_fuzzy_variables() This wont work on checkpoints!
                        rule_base.purge_rules(self.tolerance)
                        rule_base.rename_cons(self.nclasses_)
                        checkpoint_rules = rule_base.print_rules(True)
                        f.write(checkpoint_rules)     

        else:
            res = minimize(problem,
                        algorithm,
                        # termination,
                        ("n_gen", n_gen),
                        seed=random_state,
                        copy_algorithm=False,
                        save_history=False,
                        verbose=self.verbose)
        
        pop = res.pop
        fitness_last_gen = pop.get('F')
        best_solution = np.argmin(fitness_last_gen)
        best_individual = pop.get('X')[best_solution, :]

        
        self.performance = 1 - fitness_last_gen[best_solution]

        try:
            self.var_names = list(X.columns)
            self.X = X.values
        except AttributeError:
            self.X = X
            self.var_names = [str(ix) for ix in range(X.shape[1])]

        self.rule_base = problem._construct_ruleBase(
        best_individual, self.fuzzy_type)
        self.rule_base.rename_cons(self.classes_names)

        self.eval_performance = evr.evalRuleBase(
        self.rule_base, np.array(X), y)

        self.eval_performance.add_full_evaluation()  
        if self.lvs is None:
            self.rename_fuzzy_variables()
        self.eval_performance.add_classification_metrics()
        self.rule_base.purge_rules(self.tolerance)
        

    def load_master_rule_base(self, rule_base: rules.MasterRuleBase) -> None:
        '''
        Loads a master rule base to be used in the prediction process.

        :param rule_base: ruleBase object.
        :return: None
        '''
        self.rule_base = rule_base
        self.nRules = len(rule_base.get_rules())
        self.nAnts = len(rule_base.get_rules()[0].antecedents)
        self.n_class = len(rule_base)
        self.nclasses_ = rule_base.consequent_names
        

    def forward(self, X: np.array) -> np.array:
        '''
        Returns the predicted class for each sample.

        :param X: np array samples x features.
        :return: np array samples (x 1) with the predicted class.
        '''
        try:
            X = X.values  # If X was a pandas dataframe
        except AttributeError:
            pass
        
        return self.rule_base.winning_rule_predict(X)
        

    def predict(self, X: np.array) -> np.array:
        '''
        Returns the predicted class for each sample.

        :param X: np array samples x features.
        :return: np array samples (x 1) with the predicted class.
        '''
        return self.forward(X)


    def print_rules(self, return_rules:bool=False) -> None:
        '''
        Print the rules contained in the fitted rulebase.
        '''
        return self.rule_base.print_rules(return_rules)


    def plot_fuzzy_variables(self) -> None:
        '''
        Plot the fuzzy partitions in each fuzzy variable.
        '''
        fuzzy_variables = self.rule_base.rule_bases[0].antecedents

        for ix, fv in enumerate(fuzzy_variables):
            vis_rules.plot_fuzzy_variable(fv)


    def rename_fuzzy_variables(self) -> None:
        '''
        Renames the linguist labels so that high, low and so on are consistent. It does so usually after an optimization process.

        :return: None. Names are sorted accorded to the central point of the fuzzy memberships.
        '''

        for ix in range(len(self.rule_base)):
            fuzzy_variables = self.rule_base.rule_bases[ix].antecedents

            for jx, fv in enumerate(fuzzy_variables):
                # I feel so extraordinary, lifes got a hold on me...
                new_order_values = []
                possible_names = FitRuleBase.vl_names[self.n_linguist_variables]

                for zx, fuzzy_set in enumerate(fv.linguistic_variables):
                    studied_fz = fuzzy_set.type()
                    
                    if studied_fz == fs.FUZZY_SETS.temporal:
                        studied_fz = fuzzy_set.inside_type()

                    if studied_fz == fs.FUZZY_SETS.t1:
                        f1 = np.mean(
                            fuzzy_set.membership_parameters[0] + fuzzy_set.membership_parameters[1])
                    elif (studied_fz == fs.FUZZY_SETS.t2):
                        f1 = np.mean(
                            fuzzy_set.secondMF_upper[0] + fuzzy_set.secondMF_upper[1])
                    elif studied_fz == fs.FUZZY_SETS.gt2:
                        sec_memberships = fuzzy_set.secondary_memberships.values()
                        f1 = float(list(fuzzy_set.secondary_memberships.keys())[np.argmax(
                            [fzm.membership_parameters[2] for ix, fzm in enumerate(sec_memberships)])])

                    new_order_values.append(f1)

                new_order = np.argsort(np.array(new_order_values))
                fuzzy_sets_vl = fv.linguistic_variables

                for jx, x in enumerate(new_order):
                    fuzzy_sets_vl[x].name = possible_names[jx]


    def get_rulebase(self) -> list[np.array]:
        '''
        Get the rulebase obtained after fitting the classifier to the data.

        :return: a matrix format for the rulebase.
        '''
        return self.rule_base.get_rulebase_matrix()
    

    def reparametrice_loss(self, alpha:float, beta:float) -> None:
        '''
        Changes the parameters in the loss function. 

        :note: Does not check for convexity preservation. The user can play with these parameters as it wills.
        :param alpha: controls the MCC term.
        :param beta: controls the average rule size loss.
        '''
        self.alpha_ = alpha
        self.beta_ = beta



class ExploreRuleBases(Problem):
    '''
    Class to model as pymoo problem the fitting of a rulebase to a set of data given a series of candidate rules for a classification problem using Evolutionary strategies
    Supports type 1 and t2.
    '''

    def __init__(self, X: np.array, y: np.array, nRules: int, n_classes: int, candidate_rules: rules.MasterRuleBase, thread_runner: StarmapParallelization=None, tolerance:float = 0.01) -> None:
        '''
        Cosntructor method. Initializes the classifier with the number of antecedents, linguist variables and the kind of fuzzy set desired.

        :param X: np array or pandas dataframe samples x features.
        :param y: np vector containing the target classes. vector sample
        :param n_class: number of classes in the problem. If None (as default) it will be computed from the data.
        :param cancidate_rules: MasterRuleBase object. If not None, the classifier will use the rules in the object and ignore the conflicting parameters.
        '''
        try:
            self.var_names = list(X.columns)
            self.X = X.values
        except AttributeError:
            self.X = X
            self.var_names = [str(ix) for ix in range(X.shape[1])]

        self.tolerance = tolerance
        self.fuzzy_type = candidate_rules.fuzzy_type()
        self.y = y
        self.nCons = 1  # This is fixed to MISO rules.
        self.n_classes = n_classes
        self.candidate_rules = candidate_rules
        self.nRules = nRules
        self._precomputed_truth = rules.compute_antecedents_memberships(candidate_rules.get_antecedents(), X)

        self.vl_names = self.candidate_rules[0].antecedents[0].linguistic_variable_names()
        self.fuzzy_type = self.candidate_rules[0].antecedents[0].fuzzy_type()

        self.min_bounds = np.min(self.X, axis=0)
        self.max_bounds = np.max(self.X, axis=0)

        nTotalRules = len(self.candidate_rules.get_rules())
        # Each var is using or not a rule. 
        vars = {ix: Integer(bounds=[0, nTotalRules - 1]) for ix in range(self.nRules)}
        varbound = np.array([[0, nTotalRules- 1]] * self.nRules)

        nVar = len(vars.keys())
        if thread_runner is not None:
            super().__init__(
                vars=vars,
                n_var=nVar,
                n_obj=1,
                elementwise=True,
                vtype=int,
                xl=varbound[:, 0],
                xu=varbound[:, 1],
                elementwise_runner=thread_runner)
        else:
            super().__init__(
                vars=vars,
                n_var=nVar,
                n_obj=1,
                elementwise=True,
                vtype=int,
                xl=varbound[:, 0],
                xu=varbound[:, 1])


    def _construct_ruleBase(self, x: np.array, fuzzy_type: fs.FUZZY_SETS) -> rules.MasterRuleBase:
        '''
        Creates a valid rulebase from the given subject and the candidate rules.

        :param x: gen of a rulebase. type: dict.
        
        :return: a Master rulebase object.
        '''
        x = x.astype(int)
        # Get all rules and their consequents
        diff_consequents = np.arange(len(self.candidate_rules))
        
        # Choose the selected ones in the gen
        total_rules = self.candidate_rules.get_rules()
        chosen_rules = [total_rules[ix] for ix, val in enumerate(x)]
        rule_consequents = sum([[ix] * len(rule) for ix, rule in enumerate(self.candidate_rules)], [])
        chosen_rules_consequents = [rule_consequents[val] for ix, val in enumerate(x)]
        # Create a rule base for each consequent with the selected rules
        rule_list = [[] for _ in range(self.n_classes)]
        rule_bases = []
        for ix, consequent in enumerate(diff_consequents):
            for rx, rule in enumerate(chosen_rules):
                if chosen_rules_consequents[rx] == consequent:
                    rule_list[ix].append(rule)

            if len(rule_list[ix]) > 0:
                if fuzzy_type == fs.FUZZY_SETS.t1:
                    rule_base_cons = rules.RuleBaseT1(
                        self.candidate_rules[0].antecedents, rule_list[ix])
                elif fuzzy_type == fs.FUZZY_SETS.t2:
                    rule_base_cons = rules.RuleBaseT2(
                        self.candidate_rules[0].antecedents, rule_list[ix])
                elif fuzzy_type == fs.FUZZY_SETS.gt2:
                    rule_base_cons = rules.RuleBaseGT2(
                        self.candidate_rules[0].antecedents, rule_list[ix])
                    
                rule_bases.append(rule_base_cons)
            
        # Create the Master Rule Base object with the individual rule bases
        newMasterRuleBase = rules.MasterRuleBase(rule_bases, diff_consequents)    

        return newMasterRuleBase


    def _evaluate(self, x: np.array, out: dict, *args, **kwargs):
        '''
        :param x: array of train samples. x shape = features
            those features are the parameters to optimize.

        :param out: dict where the F field is the fitness. It is used from the outside.
        '''
        try:
            ruleBase = self._construct_ruleBase(x, self.fuzzy_type)

            score = self.fitness_func(ruleBase, self.X, self.y, self.tolerance, precomputed_truth=self._precomputed_truth)
            

            out["F"] = 1 - score
        except rules.RuleError:
            out["F"] = 1

    
    def fitness_func(self, ruleBase: rules.RuleBase, X:np.array, y:np.array, tolerance:float, alpha:float=0.0, beta:float=0.0, precomputed_truth=None) -> float:
        '''
        Fitness function for the optimization problem.
        :param ruleBase: RuleBase object
        :param X: array of train samples. X shape = (n_samples, n_features)
        :param y: array of train labels. y shape = (n_samples,)
        :param tolerance: float. Tolerance for the size evaluation.
        :return: float. Fitness value.
        '''
        ev_object = evr.evalRuleBase(ruleBase, X, y, precomputed_truth=precomputed_truth)
        ev_object.add_rule_weights()

        score_acc = ev_object.classification_eval()
        score_rules_size = ev_object.size_antecedents_eval(tolerance)
        score_nrules = ev_object.effective_rulesize_eval(tolerance)

        score = score_acc + score_rules_size * alpha + score_nrules * beta

        return score
    


class FitRuleBase(Problem):
    '''
    Class to model as pymoo problem the fitting of a rulebase for a classification problem using Evolutionary strategies. 
    Supports type 1 and iv fs (iv-type 2)
    '''

    def _init_optimize_vl(self, fuzzy_type: fs.FUZZY_SETS, n_linguist_variables: int, domain: list[(float, float)] = None):
        '''
        Inits the corresponding fields if no linguistic partitions were given.

        :param fuzzy type: FUZZY_SET enum type in fuzzy_sets module. The kind of fuzzy set used.
        :param n_linguistic_variables: number of linguistic variables per antecedent.
        :param domain: list of the limits for each variable. If None (default) the classifier will compute them empirically.
        '''
        self.lvs = None
        try:
            self.vl_names = FitRuleBase.vl_names[n_linguist_variables]
        except IndexError:
            self.vl_names = [str(ix) for ix in range(n_linguist_variables)]

        self.fuzzy_type = fuzzy_type
        self.n_lv_possible = n_linguist_variables
        self.domain = domain
        self._precomputed_truth = None

    def _init_precomputed_vl(self, linguist_variables: list[fs.fuzzyVariable], X: np.array):
        '''
        Inits the corresponding fields if linguistic partitions for each variable are given.

        :param linguistic_variables: list of fuzzyVariables type.
        :param X: np array samples x features.
        '''
        self.lvs = linguist_variables
        self.vl_names = self.lvs[0].linguistic_variable_names()
        self.n_lv_possible = len(self.lvs[0])
        self.fuzzy_type = self.lvs[0].fs_type
        self.domain = None
        self._precomputed_truth = rules.compute_antecedents_memberships(linguist_variables, X)

    vl_names = [  # Linguistic variable names preestated for some specific cases.
        [],
        [],
        ['Low', 'High'],
        ['Low', 'Medium', 'High'],
        ['Low', 'Medium', 'High', 'Very High'],
        ['Very Low', 'Low', 'Medium', 'High', 'Very High']
    ]

    def __init__(self, X: np.array, y: np.array, nRules: int, nAnts: int, n_classes: int, thread_runner: StarmapParallelization=None, 
                 linguistic_variables:list[fs.fuzzyVariable]=None, n_linguist_variables:int=3, fuzzy_type=fs.FUZZY_SETS.t1, domain:list=None,
                 tolerance:float=0.01, alpha:float=0.0, beta:float=0.0) -> None:
        '''
        Cosntructor method. Initializes the classifier with the number of antecedents, linguist variables and the kind of fuzzy set desired.

        :param X: np array or pandas dataframe samples x features.
        :param y: np vector containing the target classes. vector sample
        :param nRules: number of rules to optimize.
        :param nAnts: max number of antecedents to use.
        :param n_class: number of classes in the problem. If None (as default) it will be computed from the data.
        :param linguistic_variables: list of linguistic variables precomputed. If given, the rest of conflicting arguments are ignored.
        :param n_linguistic_variables: number of linguistic variables per antecedent.
        :param fuzzy_type: Define the fuzzy set or fuzzy set extension used as linguistic variable.
        :param domain: list with the upper and lower domains of each input variable. If None (as default) it will stablish the empirical min/max as the limits.
        '''
        try:
            self.var_names = list(X.columns)
            self.X = X.values
        except AttributeError:
            self.X = X
            self.var_names = [str(ix) for ix in range(X.shape[1])]

        try:
            self.tolerance = tolerance
        except KeyError:
            self.tolerance = 0.001

        self.y = y
        self.classes_names = np.unique(y)
        self.nRules = nRules
        self.nAnts = nAnts
        self.nCons = 1  # This is fixed to MISO rules.

        if n_classes is not None:
            self.n_classes = n_classes
        else:
            self.n_classes = len(np.unique(y))

        if linguistic_variables is not None:
            self._init_precomputed_vl(linguistic_variables, X)
        else:
            self._init_optimize_vl(
                fuzzy_type, n_linguist_variables)

        if self.domain is None:
            # If all the variables are numerical, then we can compute the min/max of the domain.
            if np.all([np.issubdtype(self.X[:, ix].dtype, np.number) for ix in range(self.X.shape[1])]):
                self.min_bounds = np.min(self.X, axis=0)
                self.max_bounds = np.max(self.X, axis=0)
            else:
                self.min_bounds = np.zeros(self.X.shape[1])
                self.max_bounds = np.zeros(self.X.shape[1])

                for ix in range(self.X.shape[1]):
                    if np.issubdtype(self.X[:, ix].dtype, np.number):
                        self.min_bounds[ix] = np.min(self.X[:, ix])
                        self.max_bounds[ix] = np.max(self.X[:, ix])
                    else:
                        self.min_bounds[ix] = 0
                        self.max_bounds[ix] = len(np.unique(self.X[:, ix]))
        else:
            self.min_bounds, self.max_bounds = self.domain

        self.antecedents_referencial = [np.linspace(
            self.min_bounds[ix], self.max_bounds[ix], 100) for ix in range(self.X.shape[1])]

        possible_antecedent_bounds = np.array(
            [[0, self.X.shape[1] - 1]] * self.nAnts * self.nRules)  
        vl_antecedent_bounds = np.array(
            [[-1, self.n_lv_possible - 1]] * self.nAnts * self.nRules) # -1 means not caring
        antecedent_bounds = np.concatenate(
            (possible_antecedent_bounds, vl_antecedent_bounds))
        vars_antecedent = {ix: Integer(
            bounds=antecedent_bounds[ix]) for ix in range(len(antecedent_bounds))}
        aux_counter = len(vars_antecedent)

        if self.lvs is None:
            assert self.n_lv_possible >= 1, 'At least 1 possible value for a linguistic variable is required.'

            self.feature_domain_bounds = np.array(
                [[0, 99] for ix in range(self.X.shape[1])])
            size_multiplier = 4 if self.fuzzy_type == fs.FUZZY_SETS.t1 else 8
            membership_bounds = np.concatenate(
                [self.feature_domain_bounds] * self.n_lv_possible * size_multiplier)
            vars_memeberships = {
                aux_counter + ix: Integer(bounds=membership_bounds[ix]) for ix in range(len(membership_bounds))}
            aux_counter += len(vars_memeberships)

        final_consequent_bounds = np.array(
            [[-1, self.n_classes - 1]] * self.nRules)
        vars_consequent = {aux_counter + ix: Integer(
            bounds=final_consequent_bounds[ix]) for ix in range(len(final_consequent_bounds))}

        if self.lvs is None:
            vars = {key: val for d in [
                vars_antecedent, vars_memeberships, vars_consequent] for key, val in d.items()}
            varbound = np.concatenate(
                (antecedent_bounds, membership_bounds, final_consequent_bounds), axis=0)
        else:
            vars = {key: val for d in [vars_antecedent,
                                       vars_consequent] for key, val in d.items()}
            varbound = np.concatenate(
                (antecedent_bounds, final_consequent_bounds), axis=0)

        nVar = len(varbound)
        self.single_gen_size = nVar
        self.alpha_ = alpha
        self.beta_ = beta

        if thread_runner is not None:
            super().__init__(
                vars=vars,
                n_var=nVar,
                n_obj=1,
                elementwise=True,
                vtype=int,
                xl=varbound[:, 0],
                xu=varbound[:, 1],
                elementwise_runner=thread_runner)
        else:
            super().__init__(
                vars=vars,
                n_var=nVar,
                n_obj=1,
                elementwise=True,
                vtype=int,
                xl=varbound[:, 0],
                xu=varbound[:, 1])


    def encode_rulebase(self, rule_base: rules.MasterRuleBase, optimize_lv: bool) -> np.array:
        '''
        Given a rule base, constructs the corresponding gene associated with that rule base.

        GENE STRUCTURE

        First: antecedents chosen by each rule. Size: nAnts * nRules (index of the antecedent)
        Second: Variable linguistics used. Size: nAnts * nRules
        Third: Parameters for the fuzzy partitions of the chosen variables. Size: nAnts * self.n_linguistic_variables * 8|4 (2 trapezoidal memberships if t2)
        Four: Consequent classes. Size: nRules

        :param rule_base: rule base object.
        :param optimize_lv: if True, the gene is prepared to optimize the membership functions.
        :param antecedents_referencial: list of lists. Each list contains the referencial for each variable.  Required only if optimize_lv is True.
        :return: np array of size self.single_gen_size.
        '''
        gene = np.zeros((self.single_gen_size,))

        n_lv_possible = len(rule_base.rule_bases[0].antecedents[0].linguistic_variables)
        fz_type = rule_base.fuzzy_type()
        rule_consequents = rule_base.get_consequents()
        nreal_rules = len(rule_consequents)
        mf_size = 4 if fz_type == fs.FUZZY_SETS.t1 else 8

        # Pointer to the fourth section of the gene: consequents
        if optimize_lv:
            # If lv memberships are optimized.
            fourth_pointer = 2 * self.nAnts * self.nRules + \
                len(rule_base.antecedents) * n_lv_possible * mf_size
        else:
            # If no memberships are optimized.
            fourth_pointer = 2 * self.nAnts * self.nRules

        # First and second sections of the gene: antecedents and linguistic variables
        for i0, rule in enumerate(rule_base.get_rules()):  # Reconstruct the rules
            first_pointer = i0 * self.nAnts
            second_pointer = (self.nRules * self.nAnts) + i0 * self.nAnts

            for ax, linguistic_variable in enumerate(rule.antecedents):
                gene[first_pointer + ax] = ax
                gene[second_pointer + ax] = linguistic_variable
            
            # Update the fourth section of the gene: consequents using the fourth pointer
            gene[fourth_pointer + i0] = rule_consequents[i0]

        # Fill the rest of the rules with don't care values
        nvoid_rules = self.nRules - nreal_rules
        for vx in range(nvoid_rules):
            first_pointer = nreal_rules * self.nAnts + vx * self.nAnts
            second_pointer = (self.nRules * self.nAnts) + nreal_rules * self.nAnts + vx * self.nAnts

            for ax, linguistic_variable in enumerate(rule.antecedents):
                gene[first_pointer + ax] = ax
                gene[second_pointer + ax] = -1
            
            # Update the fourth section of the gene: consequents using the fourth pointer
            gene[fourth_pointer + nreal_rules + vx] = -1

        if optimize_lv:
            # If lv memberships are optimized.
            third_pointer = 2 * self.nAnts * self.nRules
            aux_pointer = 0
            for ix, fuzzy_variable in enumerate(rule_base.get_antecedents()):
                for linguistic_variable in range(n_lv_possible):
                    fz_parameters = fuzzy_variable[linguistic_variable].membership_parameters
                    for jx, fz_parameter in enumerate(fz_parameters):
                        closest_idx = (np.abs(np.asarray(self.antecedents_referencial[ix]) - fz_parameter)).argmin()
                        gene[third_pointer + aux_pointer] = closest_idx
                        aux_pointer += 1
                    
        return np.array(list(map(int, gene)))
        


    def _construct_ruleBase(self, x: np.array, fz_type: fs.FUZZY_SETS, **kwargs) -> rules.MasterRuleBase:
        '''
        Given a subject, it creates a rulebase according to its specification.

        :param x: gen of a rulebase. type: dict.
        :param fz_type: a enum type. Check fuzzy_sets for complete specification (two fields, t1 and t2, to mark which fs you want to use)
        :return: a rulebase object.

        kwargs:
            - time_moment: if temporal fuzzy sets are used with different partitions for each time interval, 
                            then this parameter is used to specify which time moment is being used.
        '''

        rule_list = [[] for _ in range(self.n_classes)]

        mf_size = 4 if fz_type == fs.FUZZY_SETS.t1 else 8
        '''
        GEN STRUCTURE

        First: antecedents chosen by each rule. Size: nAnts * nRules
        Second: Variable linguistics used. Size: nAnts * nRules
        Third: Parameters for the fuzzy partitions of the chosen variables. Size: X.shape[1] * self.n_linguistic_variables * 8|4 (2 trapezoidal memberships if t2)
        Four: Consequent classes. Size: nRules (fixed to 4 right now)
        '''
        if self.lvs is None:
            # If memberships are optimized.
            fourth_pointer = 2 * self.nAnts * self.nRules + \
                self.X.shape[1] * self.n_lv_possible * mf_size
        else:
            # If no memberships are optimized.
            fourth_pointer = 2 * self.nAnts * self.nRules

        aux_pointer = 0
        min_domain = np.min(self.X, axis=0)
        max_domain = np.max(self.X, axis=0)

        # Integer sampling doesnt work fine in pymoo, so we do this (which is btw what pymoo is really doing if you just set integer optimization)
        try:
            # subject might come as a dict.
            x = np.array(list(x.values())).astype(int)
        except AttributeError:
            x = x.astype(int)

        for i0 in range(self.nRules):  # Reconstruct the rules
            first_pointer = i0 * self.nAnts
            chosen_ants = x[first_pointer:first_pointer + self.nAnts]

            second_pointer = (i0 * self.nAnts) + (self.nAnts * self.nRules)
            # Shape: self.nAnts + self.n_lv_possible  + 1
            antecedent_parameters = x[second_pointer:second_pointer+self.nAnts]

            init_rule_antecedents = np.zeros(
                (self.X.shape[1],)) - 1  # -1 is dont care
            for jx, ant in enumerate(chosen_ants):
                if self.lvs is not None:
                    antecedent_parameters[jx] = min(antecedent_parameters[jx], len(self.lvs[ant]) - 1)
                else:
                    antecedent_parameters[jx] = min(antecedent_parameters[jx], self.n_lv_possible - 1)

                init_rule_antecedents[ant] = antecedent_parameters[jx]

            consequent_idx = x[fourth_pointer + aux_pointer]
            assert consequent_idx < self.n_classes, "Consequent class is not valid. Something in the gene is wrong."
            aux_pointer += 1
 
            if consequent_idx != -1 and np.any(init_rule_antecedents != -1):
                rule_list[consequent_idx].append(
                    rules.RuleSimple(init_rule_antecedents, 0))

        # If we optimize the membership functions
        if self.lvs is None:
            third_pointer = 2 * self.nAnts * self.nRules
            aux_pointer = 0
            antecedents = []

            for fuzzy_variable in range(self.X.shape[1]):
                linguistic_variables = []

                for linguistic_variable in range(self.n_lv_possible):
                    parameter_pointer = third_pointer + aux_pointer
                    fz_parameters_idx = x[parameter_pointer:parameter_pointer + mf_size]
                    fz_parameters = self.antecedents_referencial[fuzzy_variable][fz_parameters_idx]
                    aux_pointer += mf_size

                    if fz_type == fs.FUZZY_SETS.t2:
                        fz_parameters[0:6] = np.sort(fz_parameters[0:6])
                        mu = [np.min(fz_parameters[0:2]), fz_parameters[2],
                              fz_parameters[3], np.max(fz_parameters[4:6])]
                        ml = [np.max(fz_parameters[0:2]), fz_parameters[2],
                              fz_parameters[3], np.min(fz_parameters[4:6])]
                        height = fz_parameters[6] / np.max(fz_parameters)

                        ivfs = fs.IVFS(self.vl_names[linguistic_variable], ml, mu,
                                       (min_domain[fuzzy_variable], max_domain[fuzzy_variable]), lower_height=height)
                    else:
                        ivfs = fs.FS(self.vl_names[linguistic_variable], np.sort(fz_parameters[0:4]),
                                     (min_domain[fuzzy_variable], max_domain[fuzzy_variable]))
                    linguistic_variables.append(ivfs)

                antecedents.append(fs.fuzzyVariable(
                    self.var_names[fuzzy_variable], linguistic_variables))
        else:
            try:
                antecedents = self.lvs[kwargs['time_moment']]
            except:
                antecedents = self.lvs

        for i in range(self.n_classes):
            if fz_type == fs.FUZZY_SETS.temporal:
                fz_type = self.lvs[0][0].inside_type()

            if fz_type == fs.FUZZY_SETS.t1:
                rule_base = rules.RuleBaseT1(antecedents, rule_list[i])
            elif fz_type == fs.FUZZY_SETS.t2:
                rule_base = rules.RuleBaseT2(antecedents, rule_list[i])
            elif fz_type == fs.FUZZY_SETS.gt2:
                rule_base = rules.RuleBaseGT2(antecedents, rule_list[i])
            

            if i == 0:
                res = rules.MasterRuleBase([rule_base], self.classes_names)
            else:
                res.add_rule_base(rule_base)

        res.rename_cons(self.classes_names)

        return res


    def _evaluate(self, x: np.array, out: dict, *args, **kwargs):
        '''
        :param x: array of train samples. x shape = features
            those features are the parameters to optimize.

        :param out: dict where the F field is the fitness. It is used from the outside.
        '''
        ruleBase = self._construct_ruleBase(x, self.fuzzy_type)

        if len(ruleBase.get_rules()) > 0:
            score = self.fitness_func(ruleBase, self.X, self.y, self.tolerance, self.alpha_, self.beta_, self._precomputed_truth)
        else:
            score = 0.0
        
        out["F"] = 1 - score
    

    def fitness_func(self, ruleBase: rules.RuleBase, X:np.array, y:np.array, tolerance:float, alpha:float=0.0, beta:float=0.0, precomputed_truth:np.array=None) -> float:
        '''
        Fitness function for the optimization problem.
        :param ruleBase: RuleBase object
        :param X: array of train samples. X shape = (n_samples, n_features)
        :param y: array of train labels. y shape = (n_samples,)
        :param tolerance: float. Tolerance for the size evaluation.
        :param alpha: float. Weight for the accuracy term.
        :param beta: float. Weight for the average rule size term.
        :param precomputed_truth: np array. If given, it will be used as the truth values for the evaluation.
        :return: float. Fitness value.
        '''
        if precomputed_truth is None:
            precomputed_truth = rules.compute_antecedents_memberships(ruleBase.antecedents, X)

        ev_object = evr.evalRuleBase(ruleBase, X, y, precomputed_truth=precomputed_truth)
        ev_object.add_full_evaluation()
        ruleBase.purge_rules(tolerance)

        if len(ruleBase.get_rules()) > 0: 
            score_acc = ev_object.classification_eval()
            score_rules_size = ev_object.size_antecedents_eval(tolerance)
            score_nrules = ev_object.effective_rulesize_eval(tolerance)

            score = score_acc + score_rules_size * alpha + score_nrules * beta
        else:
            score = 0.0
            
        return score
    




#! /usr/bin/env python
"""Toolbox for explainable AI using fuzzy logic."""
from __future__ import absolute_import

import os
# read the contents of your README file
from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

DISTNAME = "ex_fuzzy"
DESCRIPTION = "Library to perform explainable AI using fuzzy logic."
MAINTAINER = "Javier Fumanal Idocin"
MAINTAINER_EMAIL = "javierfumanalidocin@gmail.com"
URL = "https://github.com/Fuminides/ex-fuzzy"
LICENSE = "GPL-3.0"
DOWNLOAD_URL = "https://pypi.org/project/ex-fuzzy/"
VERSION = "1.1.6"
INSTALL_REQUIRES = ["numpy", "networkx", "matplotlib", "pymoo", "pandas", "scikit-learn"]
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Software Development :: Libraries",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    packages=['ex_fuzzy'],
    package_dir={'ex_fuzzy': 'ex_fuzzy/ex_fuzzy'},
    install_requires=INSTALL_REQUIRES,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=CLASSIFIERS,
)



"""
Module to perform rule mining in a pandas dataframe or numpy array. The methods use the support of the different itemsets to look for good 
rule candidates. It can be used then by a Genetic optimizator from evolutionary_fit module to search the optimal combination of them.

"""
import typing
from itertools import product, combinations

import pandas as pd
import numpy as np

try:
    from . import rules as rl
    from . import fuzzy_sets as fs
    from . import maintenance as mnt
    from . import utils
except ImportError:
    import utils
    import rules as rl
    import fuzzy_sets as fs
    import maintenance as mnt


def _generate_combinations(lists: list, k: int) -> typing.Iterator:
    '''
    Generate all the combinations between elements of different lists of length k without repeting elements of the same list.

    :param lists: list of lists.
    :param k: integer with the length of the combinations.
    :return: a list with all the combinations.
    '''
    # Get all combinations of elements for k
    all_combs = combinations(np.arange(len(lists)), k)
    
    # For those elements, get the cartesian product between them
    for comb in all_combs:
        selected_lists = [lists[x] for x in comb]
        all_combinations = product(*selected_lists)

        # Add them to the global combination list
        yield all_combinations


def rule_search(data: pd.DataFrame, fuzzy_variables: dict[fs.fuzzyVariable], support_threshold:float=0.05, max_depth:int=None) -> list:
    '''
    Computes the apriori algorithm for the given dataframe and threshold the support.
    
    :param data: Dataframe of shape: samples x features
    :param fuzzy variables: dict that maps each feature name with a fuzzy variable.
    :param support_threshold: minimum support to consider frequent an itemset.
    :return: all the frequent itemsets as a list.
    '''
    list_possible_vars = []
    for ix, fuzzy_variable in enumerate(fuzzy_variables):
        list_possible_vars.append([(ix, ax) for ax in range(len(fuzzy_variable))])

    memberships = [fuzzy_variables[ix](data.iloc[:, ix].values) for ix in range(data.shape[1])]
    freq_itemsets = []

    if max_depth is None:
        max_depth = data.shape[1]
    
    # For all possible lengths
    for r in range(max_depth):
        all_r_combs = _generate_combinations(list_possible_vars, r+1)

        # Iterate through the possible itemsets
        for itemsets in all_r_combs:
            for ix, itemset in enumerate(itemsets):
                relevant_memberships = []
                for item in itemset:
                    item_var, item_vl = item
                    relevant_memberships.append([memberships[item_var][item_vl]])
                
                array_membership = np.array(relevant_memberships).T[:,0,:]
                support = np.mean(np.min(array_membership, axis=1))
                if fuzzy_variables[0].fuzzy_type == fs.FUZZY_SETS.t2 or fuzzy_variables[0].fuzzy_type == fs.FUZZY_SETS.gt2:
                    support = np.mean(support, axis=1)

                if support > support_threshold:
                    freq_itemsets.append(itemset) 

    return freq_itemsets


def generate_rules_from_itemsets(itemsets:list, nAnts:int) -> list[rl.RuleSimple]:
    '''
    Given a list of itemsets, it creates the rules for each one and returns a list of rules containing them.

    :param itemsets: list of tuple (antecedent, linguistic variable value) 
    :param nAnts: number of possible antecedents.
    :return: the rules for ech itemset.
    '''
    rules = []
    for itemset in itemsets:
        template = np.ones((nAnts, )) * -1
        for ant, vl in itemset:
            template[ant] = vl
        
        rule = rl.RuleSimple(list(template))
        rules.append(rule)

    return rules


def mine_rulebase_support(x: pd.DataFrame, fuzzy_variables:list[fs.fuzzyVariable], support_threshold:float=0.05, max_depth:int=3) -> rl.RuleBase:
    '''
    Search the data for associations that are frequent given a list of fuzzy variables for each antecedent.

    :param x: the data to mine. Dims: samples x features.
    :param fuzzy_variables: list of the fuzzy variables for each of the input variables.
    :param support_threshold: minimum threshold to decide if prune or not the rule.
    :param max_depth: maximum number of antecedents per rule.
    :return: a rulebase object with the rules denoted as good.
    '''
    
    if mnt.save_usage_flag:
            mnt.usage_data[mnt.usage_categories.RuleMining]['mine_rulebase'] += 1
            
    freq_itemsets = rule_search(x, fuzzy_variables, support_threshold, max_depth)
    rule_list = generate_rules_from_itemsets(freq_itemsets, len(fuzzy_variables))

    fuzzy_type = fuzzy_variables[0].fs_type

    if fuzzy_type == fs.FUZZY_SETS.t1:
        rule_base = rl.RuleBaseT1(fuzzy_variables, rule_list)
    elif fuzzy_type == fs.FUZZY_SETS.t2:
        rule_base = rl.RuleBaseT2(fuzzy_variables, rule_list)
    elif fuzzy_type == fs.FUZZY_SETS.gt2:
        rule_base = rl.RuleBaseGT2(fuzzy_variables, rule_list)
    
    return rule_base


def prune_rules_confidence_lift(x: pd.DataFrame, y:np.array, rules: rl.MasterRuleBase, fuzzy_variables: list[fs.fuzzyVariable], confidence_threshold:float=0.5, 
                                lift_threshold:float=1.05):
    '''
    Removes the rules from the rule base that do not meet a minimum value for confidence and lift measures.

    Confidence is the ratio of rules that have a particular antecedent and consequent, and those that only have the antecedent.
    Lift is ratio between confidence and expected confidence, which is the percentage of class samples in the original data.

    :param x: data to mine. samples x features.
    :param y: class vector.
    :param rules: MasterRuleBase object with the rules to prune.
    :param fuzzy_variables: a list of the fuzzy variables per antecedent.
    :param confidence_threshold: minimum confidence required to the rules.
    :param lift_threshold: minimum lift required to the rules.
    '''
    for ix, rule_base in enumerate(rules):
        delete_list = []
        relevant_class = ix
        relevant_class_samples = x.loc[np.equal(y, relevant_class), :]

        for jx, rule in enumerate(rule_base):
            real_nAnts = sum([ant != -1 for ant in rule])
            global_membership_array = np.zeros((x.shape[0], real_nAnts))
            class_samples_membership_array = np.zeros((relevant_class_samples.shape[0], real_nAnts))
            ant_counter = 0
            for zx, antecedent in enumerate(rule):
                if antecedent != -1:
                    global_membership_array[:, ant_counter] = fuzzy_variables[zx](x[fuzzy_variables[zx].name])[antecedent]
                    class_samples_membership_array[:, ant_counter] = fuzzy_variables[zx](relevant_class_samples[fuzzy_variables[zx].name])[antecedent]
                    ant_counter += 1

            # Compute rule confidence
            global_support = np.mean(np.min(global_membership_array, axis=1), axis=0)
            class_support = np.mean(np.min(class_samples_membership_array, axis=1), axis=0)
            if fuzzy_variables[0].fuzzy_type == fs.FUZZY_SETS.t2 or fuzzy_variables[0].fuzzy_type == fs.FUZZY_SETS.gt2:
                    global_support = np.mean(global_support, axis=1)
                    class_support = np.mean(class_support, axis=1)

            rule_confidence = class_support / global_support
            rule_lift = rule_confidence / np.mean(np.equal(relevant_class, y))

            if rule_confidence < confidence_threshold or rule_lift < lift_threshold:
                delete_list.append(jx)

        rule_base.remove_rules(delete_list)


def simple_mine_rulebase(x: pd.DataFrame, fuzzy_type:fs.FUZZY_SETS=fs.FUZZY_SETS.t1, support_threshold:float=0.05, max_depth:int=3) -> rl.RuleBase:
    '''
    Search the data for associations that are frequent. Computes the fuzzy variables using a 3 label partition (low, medium, high).

    :param x: the data to mine. Dims: samples x features.
    :param fuzzy_type: fuzzy type to use.
    :param support_threshold: minimum threshold to decide if prune or not the rule.
    :param max_depth: maximum number of antecedents per rule.
    :return: a rulebase object with the rules denoted as good.
    '''
    
    precomputed_partitions = utils.construct_partitions(x, fuzzy_type)
    return mine_rulebase_support(x, precomputed_partitions, support_threshold, max_depth)


def multiclass_mine_rulebase(x: pd.DataFrame, y: np.array, fuzzy_variables:list[fs.fuzzyVariable], support_threshold:float=0.05, max_depth:int=3,
                             confidence_threshold:float=0.05, lift_threshold:float=1.05) -> rl.MasterRuleBase:
    '''
    Search the data for associations that are frequent and have good confidence/lift values given a list of fuzzy variables for each antecedent. Computes a different ruleBase for each 
    class and then uses them to form a MasterRuleBase.

    :param x: the data to mine. Dims: samples x features.
    :param fuzzy_variables: list of the fuzzy variables for each of the input variables.
    :param support_threshold: minimum threshold to decide if prune or not the rule.
    :param max_depth: maximum number of antecedents per rule.
    :param confidence_threshold: minimum confidence value.
    :param lift_threshold: 
    :return: a rulebase object with the rules denoted as good.
    '''
    unique_classes = np.unique(y)
    rulebases = []
    for yclass in unique_classes:
        selected_samples = np.equal(yclass, y) 
        selected_x = x.loc[selected_samples, :]

        rulebase = mine_rulebase_support(selected_x, fuzzy_variables, support_threshold, max_depth)
        rulebases.append(rulebase)

    master_rulebase = rl.MasterRuleBase(rulebases, list(map(str, unique_classes)))
    prune_rules_confidence_lift(x, y, master_rulebase, fuzzy_variables, confidence_threshold, lift_threshold)
    return master_rulebase


def simple_multiclass_mine_rulebase(x: pd.DataFrame, y: np.array, fuzzy_type:fs.FUZZY_SETS, support_threshold:float=0.05, max_depth:int=3,
                                    confidence_threshold:float=0.5, lift_threshold:float=1.1) -> rl.MasterRuleBase:
    '''
    Search the data for associations that are frequent and have good confidence/lift values given a list of fuzzy variables for each antecedent. 
    Computes a different ruleBase for each class and then uses them to form a MasterRuleBase.

    Computes the fuzzy variables using a 3 label partition (low, medium, high).

    :param x: the data to mine. Dims: samples x features.
    :param fuzzy_type: fuzzy type to use.
    :param support_threshold: minimum threshold to decide if prune or not the rule.
    :param max_depth: maximum number of antecedents per rule.
    :return: a rulebase object with the rules denoted as good.
    '''
    precomputed_partitions = utils.construct_partitions(x, fuzzy_type)
    return multiclass_mine_rulebase(x, y, precomputed_partitions, support_threshold, max_depth, 
                                    confidence_threshold=confidence_threshold, lift_threshold=lift_threshold)








"""
Functions that contain some general functions to eval already fitted fuzzy rule based models.
It can also be used to visualize rules and fuzzy partitions.
"""
import numpy as np
import pandas as pd
from sklearn.metrics import matthews_corrcoef

try:
      from . import evolutionary_fit as evf
      from . import vis_rules
except ImportError:
      import evolutionary_fit as evf
      import vis_rules


def eval_fuzzy_model(fl_classifier: evf.BaseFuzzyRulesClassifier, X_train: np.array, y_train: np.array,
                     X_test: np.array, y_test: np.array, plot_rules=True, print_rules=True, plot_partitions=True, 
                     return_rules=False, print_accuracy=True, print_matthew=True, export_path:str=None) -> None:
    '''
    Function that evaluates a fuzzy rule based model. It also plots the rules and the fuzzy partitions.

    :param fl_classifier: Fuzzy rule based model.
    :param X_train: Training data.
    :param y_train: Training labels.
    :param X_test: Test data.
    :param y_test: Test labels.
    :param plot_rules: If True, it plots the rules.
    :param print_rules: If True, it prints the rules.
    :param plot_partitions: If True, it plots the fuzzy partitions.
    :return: None
    '''
    # Get the unique classes from the classifier
    unique_classes = fl_classifier.classes_names

    # Convert the names from the labels to the corresponding class if necessary
    if isinstance(y_train[0], str):
      y_train = np.array([list(unique_classes).index(str(y)) for y in y_train])
      y_test = np.array([list(unique_classes).index(str(y)) for y in y_test])
    
    if print_accuracy:
      print('------------')
      print('ACCURACY')
      print('Train performance: ' +
            str(np.mean(np.equal(y_train, fl_classifier.predict(X_train)))))
      print('Test performance: ' +
            str(np.mean(np.equal(y_test, fl_classifier.predict(X_test)))))
      print('------------')
    if print_matthew:
      print('MATTHEW CORRCOEF')
      print('Train performance: ' +
            str(matthews_corrcoef(y_train, fl_classifier.predict(X_train))))
      print('Test performance: ' +
            str(matthews_corrcoef(y_test, fl_classifier.predict(X_test))))
      print('------------')


    if print_rules or return_rules:
        res = fl_classifier.print_rules(True)

        if print_rules:
            print(res)

    if plot_partitions:
        fl_classifier.plot_fuzzy_variables()

    if plot_rules:
        vis_rules.visualize_rulebase(fl_classifier.rule_base, export_path=export_path)
        
    if return_rules:
        return res
    
    
    